from __future__ import annotations

from collections.abc import Mapping

import betterproto

from . import (
    chat,
    clan,
    comments,
    content_server,
    econ,
    friend_messages,
    game_servers,
    inventory,
    parental,
    player,
    published_file,
    reviews,
    two_factor,
)

__all__ = ("UMS",)


UMS: Mapping[str, type[betterproto.Message]] = {
    "ChatRoom.CreateChatRoomGroup#1_Request": chat.CreateChatRoomGroupRequest,
    "ChatRoom.CreateChatRoomGroup#1_Response": chat.CreateChatRoomGroupResponse,
    "ChatRoom.SaveChatRoomGroup#1_Request": chat.SaveChatRoomGroupRequest,
    "ChatRoom.SaveChatRoomGroup#1_Response": chat.SaveChatRoomGroupResponse,
    "ChatRoom.RenameChatRoomGroup#1_Request": chat.RenameChatRoomGroupRequest,
    "ChatRoom.RenameChatRoomGroup#1_Response": chat.RenameChatRoomGroupResponse,
    "ChatRoom.SetChatRoomGroupTagline#1_Request": chat.SetChatRoomGroupTaglineRequest,
    "ChatRoom.SetChatRoomGroupTagline#1_Response": chat.SetChatRoomGroupTaglineResponse,
    "ChatRoom.SetChatRoomGroupAvatar#1_Request": chat.SetChatRoomGroupAvatarRequest,
    "ChatRoom.SetChatRoomGroupAvatar#1_Response": chat.SetChatRoomGroupAvatarResponse,
    "ChatRoom.SetChatRoomGroupWatchingBroadcast#1_Request": chat.SetChatRoomGroupWatchingBroadcastRequest,
    "ChatRoom.SetChatRoomGroupWatchingBroadcast#1_Response": chat.SetChatRoomGroupWatchingBroadcastResponse,
    "ChatRoom.JoinMiniGameForChatRoomGroup#1_Request": chat.JoinMiniGameForChatRoomGroupRequest,
    "ChatRoom.JoinMiniGameForChatRoomGroup#1_Response": chat.JoinMiniGameForChatRoomGroupResponse,
    "ChatRoom.EndMiniGameForChatRoomGroup#1_Request": chat.EndMiniGameForChatRoomGroupRequest,
    "ChatRoom.EndMiniGameForChatRoomGroup#1_Response": chat.EndMiniGameForChatRoomGroupResponse,
    "ChatRoom.MuteUserInGroup#1_Request": chat.MuteUserRequest,
    "ChatRoom.MuteUserInGroup#1_Response": chat.MuteUserResponse,
    "ChatRoom.KickUserFromGroup#1_Request": chat.KickUserRequest,
    "ChatRoom.KickUserFromGroup#1_Response": chat.KickUserResponse,
    "ChatRoom.SetUserBanState#1_Request": chat.SetUserBanStateRequest,
    "ChatRoom.SetUserBanState#1_Response": chat.SetUserBanStateResponse,
    "ChatRoom.RevokeInviteToGroup#1_Request": chat.RevokeInviteRequest,
    "ChatRoom.RevokeInviteToGroup#1_Response": chat.RevokeInviteResponse,
    "ChatRoom.CreateRole#1_Request": chat.CreateRoleRequest,
    "ChatRoom.CreateRole#1_Response": chat.CreateRoleResponse,
    "ChatRoom.GetRoles#1_Request": chat.GetRolesRequest,
    "ChatRoom.GetRoles#1_Response": chat.GetRolesResponse,
    "ChatRoom.RenameRole#1_Request": chat.RenameRoleRequest,
    "ChatRoom.RenameRole#1_Response": chat.RenameRoleResponse,
    "ChatRoom.ReorderRole#1_Request": chat.ReorderRoleRequest,
    "ChatRoom.ReorderRole#1_Response": chat.ReorderRoleResponse,
    "ChatRoom.DeleteRole#1_Request": chat.DeleteRoleRequest,
    "ChatRoom.DeleteRole#1_Response": chat.DeleteRoleResponse,
    "ChatRoom.GetRoleActions#1_Request": chat.GetRoleActionsRequest,
    "ChatRoom.GetRoleActions#1_Response": chat.GetRoleActionsResponse,
    "ChatRoom.ReplaceRoleActions#1_Request": chat.ReplaceRoleActionsRequest,
    "ChatRoom.ReplaceRoleActions#1_Response": chat.ReplaceRoleActionsResponse,
    "ChatRoom.AddRoleToUser#1_Request": chat.AddRoleToUserRequest,
    "ChatRoom.AddRoleToUser#1_Response": chat.AddRoleToUserResponse,
    "ChatRoom.GetRolesForUser#1_Request": chat.GetRolesForUserRequest,
    "ChatRoom.GetRolesForUser#1_Response": chat.GetRolesForUserResponse,
    "ChatRoom.DeleteRoleFromUser#1_Request": chat.DeleteRoleFromUserRequest,
    "ChatRoom.DeleteRoleFromUser#1_Response": chat.DeleteRoleFromUserResponse,
    "ChatRoom.JoinChatRoomGroup#1_Request": chat.JoinChatRoomGroupRequest,
    "ChatRoom.JoinChatRoomGroup#1_Response": chat.JoinChatRoomGroupResponse,
    "ChatRoom.InviteFriendToChatRoomGroup#1_Request": chat.InviteFriendToChatRoomGroupRequest,
    "ChatRoom.InviteFriendToChatRoomGroup#1_Response": chat.InviteFriendToChatRoomGroupResponse,
    "ChatRoom.LeaveChatRoomGroup#1_Request": chat.LeaveChatRoomGroupRequest,
    "ChatRoom.LeaveChatRoomGroup#1_Response": chat.LeaveChatRoomGroupResponse,
    "ChatRoom.CreateChatRoom#1_Request": chat.CreateChatRoomRequest,
    "ChatRoom.CreateChatRoom#1_Response": chat.CreateChatRoomResponse,
    "ChatRoom.DeleteChatRoom#1_Request": chat.DeleteChatRoomRequest,
    "ChatRoom.DeleteChatRoom#1_Response": chat.DeleteChatRoomResponse,
    "ChatRoom.RenameChatRoom#1_Request": chat.RenameChatRoomRequest,
    "ChatRoom.RenameChatRoom#1_Response": chat.RenameChatRoomResponse,
    "ChatRoom.ReorderChatRoom#1_Request": chat.ReorderChatRoomRequest,
    "ChatRoom.ReorderChatRoom#1_Response": chat.ReorderChatRoomResponse,
    "ChatRoom.SendChatMessage#1_Request": chat.SendChatMessageRequest,
    "ChatRoom.SendChatMessage#1_Response": chat.SendChatMessageResponse,
    "ChatRoom.JoinVoiceChat#1_Request": chat.JoinVoiceChatRequest,
    "ChatRoom.JoinVoiceChat#1_Response": chat.JoinVoiceChatResponse,
    "ChatRoom.LeaveVoiceChat#1_Request": chat.LeaveVoiceChatRequest,
    "ChatRoom.LeaveVoiceChat#1_Response": chat.LeaveVoiceChatResponse,
    "ChatRoom.GetMessageHistory#1_Request": chat.GetMessageHistoryRequest,
    "ChatRoom.GetMessageHistory#1_Response": chat.GetMessageHistoryResponse,
    "ChatRoom.GetMyChatRoomGroups#1_Request": chat.GetMyChatRoomGroupsRequest,
    "ChatRoom.GetMyChatRoomGroups#1_Response": chat.GetMyChatRoomGroupsResponse,
    "ChatRoom.GetChatRoomGroupState#1_Request": chat.GetChatRoomGroupStateRequest,
    "ChatRoom.GetChatRoomGroupState#1_Response": chat.GetChatRoomGroupStateResponse,
    "ChatRoom.GetChatRoomGroupSummary#1_Request": chat.GetChatRoomGroupSummaryRequest,
    "ChatRoom.GetChatRoomGroupSummary#1_Response": chat.GetChatRoomGroupSummaryResponse,
    "ChatRoom.SetAppChatRoomGroupForceActive#1_Request": chat.SetAppChatRoomGroupForceActiveRequest,
    "ChatRoom.SetAppChatRoomGroupForceActive#1_Response": chat.SetAppChatRoomGroupForceActiveResponse,
    "ChatRoom.SetAppChatRoomGroupStopForceActive#1_Request": chat.SetAppChatRoomGroupStopForceActiveNotification,
    "ChatRoom.AckChatMessage#1_Request": chat.AckChatMessageNotification,
    "ChatRoom.CreateInviteLink#1_Request": chat.CreateInviteLinkRequest,
    "ChatRoom.CreateInviteLink#1_Response": chat.CreateInviteLinkResponse,
    "ChatRoom.GetInviteLinkInfo#1_Request": chat.GetInviteLinkInfoRequest,
    "ChatRoom.GetInviteLinkInfo#1_Response": chat.GetInviteLinkInfoResponse,
    "ChatRoom.GetInviteInfo#1_Request": chat.GetInviteInfoRequest,
    "ChatRoom.GetInviteInfo#1_Response": chat.GetInviteInfoResponse,
    "ChatRoom.GetInviteLinksForGroup#1_Request": chat.GetInviteLinksForGroupRequest,
    "ChatRoom.GetInviteLinksForGroup#1_Response": chat.GetInviteLinksForGroupResponse,
    "ChatRoom.GetBanList#1_Request": chat.GetBanListRequest,
    "ChatRoom.GetBanList#1_Response": chat.GetBanListResponse,
    "ChatRoom.GetInviteList#1_Request": chat.GetInviteListRequest,
    "ChatRoom.GetInviteList#1_Response": chat.GetInviteListResponse,
    "ChatRoom.DeleteInviteLink#1_Request": chat.DeleteInviteLinkRequest,
    "ChatRoom.DeleteInviteLink#1_Response": chat.DeleteInviteLinkResponse,
    "ChatRoom.SetSessionActiveChatRoomGroups#1_Request": chat.SetSessionActiveChatRoomGroupsRequest,
    "ChatRoom.SetSessionActiveChatRoomGroups#1_Response": chat.SetSessionActiveChatRoomGroupsResponse,
    "ChatRoom.SetUserChatGroupPreferences#1_Request": chat.SetUserChatGroupPreferencesRequest,
    "ChatRoom.SetUserChatGroupPreferences#1_Response": chat.SetUserChatGroupPreferencesResponse,
    "ChatRoom.DeleteChatMessages#1_Request": chat.DeleteChatMessagesRequest,
    "ChatRoom.DeleteChatMessages#1_Response": chat.DeleteChatMessagesResponse,
    "ChatRoom.UpdateMemberListView#1_Request": chat.UpdateMemberListViewNotification,
    "ChatRoom.SearchMembers#1_Request": chat.SearchMembersRequest,
    "ChatRoom.SearchMembers#1_Response": chat.SearchMembersResponse,
    "ChatRoom.UpdateMessageReaction#1_Request": chat.UpdateMessageReactionRequest,
    "ChatRoom.UpdateMessageReaction#1_Response": chat.UpdateMessageReactionResponse,
    "ChatRoom.GetMessageReactionReactors#1_Request": chat.GetMessageReactionReactorsRequest,
    "ChatRoom.GetMessageReactionReactors#1_Response": chat.GetMessageReactionReactorsResponse,
    "Clan.RespondToClanInvite#1_Request": clan.RespondToClanInviteRequest,
    "Clan.RespondToClanInvite#1_Response": clan.RespondToClanInviteResponse,
    "Clan.GetAdjacentPartnerEvents#1_Request": clan.GetAdjacentPartnerEventsRequest,
    "Clan.GetAdjacentPartnerEvents#1_Response": clan.GetAdjacentPartnerEventsResponse,
    "Clan.GetEventDetails#1_Request": clan.GetEventDetailsRequest,
    "Clan.GetEventDetails#1_Response": clan.GetEventDetailsResponse,
    "Clan.GetSinglePartnerEvent#1_Request": clan.GetSinglePartnerEventRequest,
    "Clan.GetSinglePartnerEvent#1_Response": clan.GetSinglePartnerEventResponse,
    "ClanChatRooms.GetClanChatRoomInfo#1_Request": chat.GetClanChatRoomInfoRequest,
    "ClanChatRooms.GetClanChatRoomInfo#1_Response": chat.GetClanChatRoomInfoResponse,
    "ClanChatRooms.SetClanChatRoomPrivate#1_Request": chat.SetClanChatRoomPrivateRequest,
    "ClanChatRooms.SetClanChatRoomPrivate#1_Response": chat.SetClanChatRoomPrivateResponse,
    "ChatRoomClient.NotifyIncomingChatMessage#1_Request": chat.IncomingChatMessageNotification,
    "ChatRoomClient.NotifyChatMessageModified#1_Request": chat.ChatMessageModifiedNotification,
    "ChatRoomClient.NotifyMemberStateChange#1_Request": chat.MemberStateChangeNotification,
    "ChatRoomClient.NotifyChatRoomHeaderStateChange#1_Request": chat.ChatRoomHeaderStateNotification,
    "ChatRoomClient.NotifyChatRoomGroupRoomsChange#1_Request": chat.ChatRoomGroupRoomsChangeNotification,
    "ChatRoomClient.NotifyShouldRejoinChatRoomVoiceChat#1_Request": chat.NotifyShouldRejoinChatRoomVoiceChatNotification,
    "ChatRoomClient.NotifyChatGroupUserStateChanged#1_Request": chat.NotifyChatGroupUserStateChangedNotification,
    "ChatRoomClient.NotifyAckChatMessageEcho#1_Request": chat.AckChatMessageNotification,
    "ChatRoomClient.NotifyChatRoomDisconnect#1_Request": chat.NotifyChatRoomDisconnectNotification,
    "ChatRoomClient.NotifyMemberListViewUpdated#1_Request": chat.MemberListViewUpdatedNotification,
    "ChatRoomClient.NotifyMessageReaction#1_Request": chat.MessageReactionNotification,
    "ContentServerDirectory.GetServersForSteamPipe#1_Request": content_server.GetServersForSteamPipeRequest,
    "ContentServerDirectory.GetServersForSteamPipe#1_Response": content_server.GetServersForSteamPipeResponse,
    "ContentServerDirectory.GetDepotPatchInfo#1_Request": content_server.GetDepotPatchInfoRequest,
    "ContentServerDirectory.GetDepotPatchInfo#1_Response": content_server.GetDepotPatchInfoResponse,
    "ContentServerDirectory.GetClientUpdateHosts#1_Request": content_server.GetClientUpdateHostsRequest,
    "ContentServerDirectory.GetClientUpdateHosts#1_Response": content_server.GetClientUpdateHostsResponse,
    "Community.GetCommentThread#1_Request": comments.GetCommentThreadRequest,
    "Community.GetCommentThread#1_Response": comments.GetCommentThreadResponse,
    "Community.PostCommentToThread#1_Request": comments.PostCommentToThreadRequest,
    "Community.PostCommentToThread#1_Response": comments.PostCommentToThreadResponse,
    "Community.DeleteCommentFromThread#1_Request": comments.DeleteCommentFromThreadRequest,
    "Community.DeleteCommentFromThread#1_Response": comments.DeleteCommentFromThreadResponse,
    "Community.RateCommentThread#1_Request": comments.RateCommentThreadRequest,
    "Community.RateCommentThread#1_Response": comments.RateCommentThreadResponse,
    "Community.GetCommentThreadRatings#1_Request": comments.GetCommentThreadRatingsRequest,
    "Community.GetCommentThreadRatings#1_Response": comments.GetCommentThreadRatingsResponse,
    "Community.RateClanAnnouncement#1_Request": comments.RateClanAnnouncementRequest,
    "Community.RateClanAnnouncement#1_Response": comments.RateClanAnnouncementResponse,
    "Econ.GetTradeOfferAccessToken#1_Request": econ.GetTradeOfferAccessTokenRequest,
    "Econ.GetTradeOfferAccessToken#1_Response": econ.GetTradeOfferAccessTokenResponse,
    "Econ.ClientGetItemShopOverlayAuthURL#1_Request": econ.GetItemShopOverlayAuthUrlRequest,
    "Econ.ClientGetItemShopOverlayAuthURL#1_Response": econ.GetItemShopOverlayAuthUrlResponse,
    "Econ.GetAssetClassInfo#1_Request": econ.GetAssetClassInfoRequest,
    "Econ.GetAssetClassInfo#1_Response": econ.GetAssetClassInfoResponse,
    "FriendMessages.GetRecentMessages#1_Request": friend_messages.GetRecentMessagesRequest,
    "FriendMessages.GetRecentMessages#1_Response": friend_messages.GetRecentMessagesResponse,
    "FriendMessages.GetActiveMessageSessions#1_Request": friend_messages.GetActiveMessageSessionsRequest,
    "FriendMessages.GetActiveMessageSessions#1_Response": friend_messages.GetActiveMessageSessionsResponse,
    "FriendMessages.SendMessage#1_Request": friend_messages.SendMessageRequest,
    "FriendMessages.SendMessage#1_Response": friend_messages.SendMessageResponse,
    "FriendMessages.AckMessage#1_Request": friend_messages.AckMessageNotification,
    "FriendMessages.UpdateMessageReaction#1_Request": friend_messages.UpdateMessageReactionRequest,
    "FriendMessages.UpdateMessageReaction#1_Response": friend_messages.UpdateMessageReactionResponse,
    "FriendMessagesClient.IncomingMessage#1_Request": friend_messages.IncomingMessageNotification,
    "FriendMessagesClient.NotifyAckMessageEcho#1_Request": friend_messages.AckMessageNotification,
    "FriendMessagesClient.MessageReaction#1_Request": friend_messages.MessageReactionNotification,
    "GameServers.GetServerList#1_Request": game_servers.GetServerListRequest,
    "GameServers.GetServerList#1_Response": game_servers.GetServerListResponse,
    "GameServers.GetServerSteamIDsByIP#1_Request": game_servers.GetServerSteamIDsByIpRequest,
    "GameServers.GetServerSteamIDsByIP#1_Response": game_servers.IPsWithSteamIDsResponse,
    "GameServers.GetServerIPsBySteamID#1_Request": game_servers.GetServerIPsBySteamIdRequest,
    "GameServers.GetServerIPsBySteamID#1_Response": game_servers.IPsWithSteamIDsResponse,
    "Inventory.GetInventory#1_Request": inventory.GetInventoryRequest,
    "Inventory.GetInventory#1_Response": inventory.Response,
    "Inventory.ExchangeItem#1_Request": inventory.ExchangeItemRequest,
    "Inventory.ExchangeItem#1_Response": inventory.Response,
    "Inventory.GetEligiblePromoItemDefIDs#1_Request": inventory.GetEligiblePromoItemDefIDsRequest,
    "Inventory.GetEligiblePromoItemDefIDs#1_Response": inventory.GetEligiblePromoItemDefIDsResponse,
    "Inventory.AddPromoItem#1_Request": inventory.AddItemRequest,
    "Inventory.AddPromoItem#1_Response": inventory.Response,
    "Inventory.SafeModifyItems#1_Request": inventory.ModifyItemsRequest,
    "Inventory.SafeModifyItems#1_Response": inventory.Response,
    "Inventory.ConsumePlaytime#1_Request": inventory.ConsumePlaytimeRequest,
    "Inventory.ConsumePlaytime#1_Response": inventory.Response,
    "Inventory.ConsumeItem#1_Request": inventory.ConsumeItemRequest,
    "Inventory.ConsumeItem#1_Response": inventory.Response,
    "Inventory.DevGenerateItem#1_Request": inventory.AddItemRequest,
    "Inventory.DevGenerateItem#1_Response": inventory.Response,
    "Inventory.DevSetNextDrop#1_Request": inventory.DevSetNextDropRequest,
    "Inventory.DevSetNextDrop#1_Response": inventory.Response,
    "Inventory.SplitItemStack#1_Request": inventory.SplitItemStackRequest,
    "Inventory.SplitItemStack#1_Response": inventory.Response,
    "Inventory.CombineItemStacks#1_Request": inventory.CombineItemStacksRequest,
    "Inventory.CombineItemStacks#1_Response": inventory.Response,
    "Inventory.GetItemDefMeta#1_Request": inventory.GetItemDefMetaRequest,
    "Inventory.GetItemDefMeta#1_Response": inventory.GetItemDefMetaResponse,
    "Inventory.GetUserPurchaseInfo#1_Request": inventory.GetUserPurchaseInfoRequest,
    "Inventory.GetUserPurchaseInfo#1_Response": inventory.GetUserPurchaseInfoResponse,
    "Inventory.PurchaseInit#1_Request": inventory.PurchaseInitRequest,
    "Inventory.PurchaseInit#1_Response": inventory.PurchaseInitResponse,
    "Inventory.PurchaseFinalize#1_Request": inventory.PurchaseFinalizeRequest,
    "Inventory.PurchaseFinalize#1_Response": inventory.Response,
    "Inventory.InspectItem#1_Request": inventory.InspectItemRequest,
    "Inventory.InspectItem#1_Response": inventory.Response,
    "InventoryClient.NotifyNewItems#1_Request": inventory.ClientNewItemsNotification,
    "Parental.EnableParentalSettings#1_Request": parental.EnableParentalSettingsRequest,
    "Parental.EnableParentalSettings#1_Response": parental.EnableParentalSettingsResponse,
    "Parental.DisableParentalSettings#1_Request": parental.DisableParentalSettingsRequest,
    "Parental.DisableParentalSettings#1_Response": parental.DisableParentalSettingsResponse,
    "Parental.GetParentalSettings#1_Request": parental.GetParentalSettingsRequest,
    "Parental.GetParentalSettings#1_Response": parental.GetParentalSettingsResponse,
    "Parental.GetSignedParentalSettings#1_Request": parental.GetSignedParentalSettingsRequest,
    "Parental.GetSignedParentalSettings#1_Response": parental.GetSignedParentalSettingsResponse,
    "Parental.SetParentalSettings#1_Request": parental.SetParentalSettingsRequest,
    "Parental.SetParentalSettings#1_Response": parental.SetParentalSettingsResponse,
    "Parental.ValidateToken#1_Request": parental.ValidateTokenRequest,
    "Parental.ValidateToken#1_Response": parental.ValidateTokenResponse,
    "Parental.ValidatePassword#1_Request": parental.ValidatePasswordRequest,
    "Parental.ValidatePassword#1_Response": parental.ValidatePasswordResponse,
    "Parental.LockClient#1_Request": parental.LockClientRequest,
    "Parental.LockClient#1_Response": parental.LockClientResponse,
    "Parental.RequestRecoveryCode#1_Request": parental.RequestRecoveryCodeRequest,
    "Parental.RequestRecoveryCode#1_Response": parental.RequestRecoveryCodeResponse,
    "Parental.DisableWithRecoveryCode#1_Request": parental.DisableWithRecoveryCodeRequest,
    "Parental.DisableWithRecoveryCode#1_Response": parental.DisableWithRecoveryCodeResponse,
    "Player.GetMutualFriendsForIncomingInvites#1_Request": player.GetMutualFriendsForIncomingInvitesRequest,
    "Player.GetMutualFriendsForIncomingInvites#1_Response": player.GetMutualFriendsForIncomingInvitesResponse,
    "Player.GetOwnedGames#1_Request": player.GetOwnedGamesRequest,
    "Player.GetOwnedGames#1_Response": player.GetOwnedGamesResponse,
    "Player.GetPlayNext#1_Request": player.GetPlayNextRequest,
    "Player.GetPlayNext#1_Response": player.GetPlayNextResponse,
    "Player.GetFriendsGameplayInfo#1_Request": player.GetFriendsGameplayInfoRequest,
    "Player.GetFriendsGameplayInfo#1_Response": player.GetFriendsGameplayInfoResponse,
    "Player.GetFriendsAppsActivity#1_Request": player.GetFriendsAppsActivityRequest,
    "Player.GetFriendsAppsActivity#1_Response": player.GetFriendsAppsActivityResponse,
    "Player.GetGameBadgeLevels#1_Request": player.GetGameBadgeLevelsRequest,
    "Player.GetGameBadgeLevels#1_Response": player.GetGameBadgeLevelsResponse,
    "Player.GetProfileBackground#1_Request": player.GetProfileBackgroundRequest,
    "Player.GetProfileBackground#1_Response": player.GetProfileBackgroundResponse,
    "Player.SetProfileBackground#1_Request": player.SetProfileBackgroundRequest,
    "Player.SetProfileBackground#1_Response": player.SetProfileBackgroundResponse,
    "Player.GetMiniProfileBackground#1_Request": player.GetMiniProfileBackgroundRequest,
    "Player.GetMiniProfileBackground#1_Response": player.GetMiniProfileBackgroundResponse,
    "Player.SetMiniProfileBackground#1_Request": player.SetMiniProfileBackgroundRequest,
    "Player.SetMiniProfileBackground#1_Response": player.SetMiniProfileBackgroundResponse,
    "Player.GetAvatarFrame#1_Request": player.GetAvatarFrameRequest,
    "Player.GetAvatarFrame#1_Response": player.GetAvatarFrameResponse,
    "Player.SetAvatarFrame#1_Request": player.SetAvatarFrameRequest,
    "Player.SetAvatarFrame#1_Response": player.SetAvatarFrameResponse,
    "Player.GetAnimatedAvatar#1_Request": player.GetAnimatedAvatarRequest,
    "Player.GetAnimatedAvatar#1_Response": player.GetAnimatedAvatarResponse,
    "Player.SetAnimatedAvatar#1_Request": player.SetAnimatedAvatarRequest,
    "Player.SetAnimatedAvatar#1_Response": player.SetAnimatedAvatarResponse,
    "Player.GetProfileItemsOwned#1_Request": player.GetProfileItemsOwnedRequest,
    "Player.GetProfileItemsOwned#1_Response": player.GetProfileItemsOwnedResponse,
    "Player.GetProfileItemsEquipped#1_Request": player.GetProfileItemsEquippedRequest,
    "Player.GetProfileItemsEquipped#1_Response": player.GetProfileItemsEquippedResponse,
    "Player.SetEquippedProfileItemFlags#1_Request": player.SetEquippedProfileItemFlagsRequest,
    "Player.SetEquippedProfileItemFlags#1_Response": player.SetEquippedProfileItemFlagsResponse,
    "Player.GetEmoticonList#1_Request": player.GetEmoticonListRequest,
    "Player.GetEmoticonList#1_Response": player.GetEmoticonListResponse,
    "Player.GetAchievementsProgress#1_Request": player.GetAchievementsProgressRequest,
    "Player.GetAchievementsProgress#1_Response": player.GetAchievementsProgressResponse,
    "Player.GetFavoriteBadge#1_Request": player.GetFavoriteBadgeRequest,
    "Player.GetFavoriteBadge#1_Response": player.GetFavoriteBadgeResponse,
    "Player.SetFavoriteBadge#1_Request": player.SetFavoriteBadgeRequest,
    "Player.SetFavoriteBadge#1_Response": player.SetFavoriteBadgeResponse,
    "Player.GetProfileCustomization#1_Request": player.GetProfileCustomizationRequest,
    "Player.GetProfileCustomization#1_Response": player.GetProfileCustomizationResponse,
    "Player.GetPurchasedProfileCustomizations#1_Request": player.GetPurchasedProfileCustomizationsRequest,
    "Player.GetPurchasedProfileCustomizations#1_Response": player.GetPurchasedProfileCustomizationsResponse,
    "Player.GetPurchasedAndUpgradedProfileCustomizations#1_Request": player.GetPurchasedAndUpgradedProfileCustomizationsRequest,
    "Player.GetPurchasedAndUpgradedProfileCustomizations#1_Response": player.GetPurchasedAndUpgradedProfileCustomizationsResponse,
    "Player.GetProfileThemesAvailable#1_Request": player.GetProfileThemesAvailableRequest,
    "Player.GetProfileThemesAvailable#1_Response": player.GetProfileThemesAvailableResponse,
    "Player.SetProfileTheme#1_Request": player.SetProfileThemeRequest,
    "Player.SetProfileTheme#1_Response": player.SetProfileThemeResponse,
    "Player.SetProfilePreferences#1_Request": player.SetProfilePreferencesRequest,
    "Player.SetProfilePreferences#1_Response": player.SetProfilePreferencesResponse,
    "Player.PostStatusToFriends#1_Request": player.PostStatusToFriendsRequest,
    "Player.PostStatusToFriends#1_Response": player.PostStatusToFriendsResponse,
    "Player.GetPostedStatus#1_Request": player.GetPostedStatusRequest,
    "Player.GetPostedStatus#1_Response": player.GetPostedStatusResponse,
    "Player.DeletePostedStatus#1_Request": player.DeletePostedStatusRequest,
    "Player.DeletePostedStatus#1_Response": player.DeletePostedStatusResponse,
    "Player.ClientGetLastPlayedTimes#1_Request": player.GetLastPlayedTimesRequest,
    "Player.ClientGetLastPlayedTimes#1_Response": player.GetLastPlayedTimesResponse,
    "Player.AcceptSSA#1_Request": player.AcceptSsaRequest,
    "Player.AcceptSSA#1_Response": player.AcceptSsaResponse,
    "Player.GetNicknameList#1_Request": player.GetNicknameListRequest,
    "Player.GetNicknameList#1_Response": player.GetNicknameListResponse,
    "Player.GetPerFriendPreferences#1_Request": player.GetPerFriendPreferencesRequest,
    "Player.GetPerFriendPreferences#1_Response": player.GetPerFriendPreferencesResponse,
    "Player.SetPerFriendPreferences#1_Request": player.SetPerFriendPreferencesRequest,
    "Player.SetPerFriendPreferences#1_Response": player.SetPerFriendPreferencesResponse,
    "Player.AddFriend#1_Request": player.AddFriendRequest,
    "Player.AddFriend#1_Response": player.AddFriendResponse,
    "Player.RemoveFriend#1_Request": player.RemoveFriendRequest,
    "Player.RemoveFriend#1_Response": player.RemoveFriendResponse,
    "Player.IgnoreFriend#1_Request": player.IgnoreFriendRequest,
    "Player.IgnoreFriend#1_Response": player.IgnoreFriendResponse,
    "Player.GetCommunityPreferences#1_Request": player.GetCommunityPreferencesRequest,
    "Player.GetCommunityPreferences#1_Response": player.GetCommunityPreferencesResponse,
    "Player.SetCommunityPreferences#1_Request": player.SetCommunityPreferencesRequest,
    "Player.SetCommunityPreferences#1_Response": player.SetCommunityPreferencesResponse,
    "Player.GetTextFilterWords#1_Request": player.GetTextFilterWordsRequest,
    "Player.GetTextFilterWords#1_Response": player.GetTextFilterWordsResponse,
    "Player.GetNewSteamAnnouncementState#1_Request": player.GetNewSteamAnnouncementStateRequest,
    "Player.GetNewSteamAnnouncementState#1_Response": player.GetNewSteamAnnouncementStateResponse,
    "Player.UpdateSteamAnnouncementLastRead#1_Request": player.UpdateSteamAnnouncementLastReadRequest,
    "Player.UpdateSteamAnnouncementLastRead#1_Response": player.UpdateSteamAnnouncementLastReadResponse,
    "Player.GetPrivacySettings#1_Request": player.GetPrivacySettingsRequest,
    "Player.GetPrivacySettings#1_Response": player.GetPrivacySettingsResponse,
    "Player.GetDurationControl#1_Request": player.GetDurationControlRequest,
    "Player.GetDurationControl#1_Response": player.GetDurationControlResponse,
    "PublishedFile.Subscribe#1_Request": published_file.SubscribeRequest,
    "PublishedFile.Subscribe#1_Response": published_file.SubscribeResponse,
    "PublishedFile.Unsubscribe#1_Request": published_file.UnsubscribeRequest,
    "PublishedFile.Unsubscribe#1_Response": published_file.UnsubscribeResponse,
    "PublishedFile.CanSubscribe#1_Request": published_file.CanSubscribeRequest,
    "PublishedFile.CanSubscribe#1_Response": published_file.CanSubscribeResponse,
    "PublishedFile.Publish#1_Request": published_file.PublishRequest,
    "PublishedFile.Publish#1_Response": published_file.PublishResponse,
    "PublishedFile.GetDetails#1_Request": published_file.GetDetailsRequest,
    "PublishedFile.GetDetails#1_Response": published_file.GetDetailsResponse,
    "PublishedFile.GetItemInfo#1_Request": published_file.GetItemInfoRequest,
    "PublishedFile.GetItemInfo#1_Response": published_file.GetItemInfoResponse,
    "PublishedFile.GetUserFiles#1_Request": published_file.GetUserFilesRequest,
    "PublishedFile.GetUserFiles#1_Response": published_file.GetUserFilesResponse,
    "PublishedFile.AreFilesInSubscriptionList#1_Request": published_file.AreFilesInSubscriptionListRequest,
    "PublishedFile.AreFilesInSubscriptionList#1_Response": published_file.AreFilesInSubscriptionListResponse,
    "PublishedFile.Update#1_Request": published_file.UpdateRequest,
    "PublishedFile.Update#1_Response": published_file.UpdateResponse,
    "PublishedFile.GetChangeHistoryEntry#1_Request": published_file.GetChangeHistoryEntryRequest,
    "PublishedFile.GetChangeHistoryEntry#1_Response": published_file.GetChangeHistoryEntryResponse,
    "PublishedFile.GetChangeHistory#1_Request": published_file.GetChangeHistoryRequest,
    "PublishedFile.GetChangeHistory#1_Response": published_file.GetChangeHistoryResponse,
    "PublishedFile.RefreshVotingQueue#1_Request": published_file.RefreshVotingQueueRequest,
    "PublishedFile.RefreshVotingQueue#1_Response": published_file.RefreshVotingQueueResponse,
    "PublishedFile.QueryFiles#1_Request": published_file.QueryFilesRequest,
    "PublishedFile.QueryFiles#1_Response": published_file.QueryFilesResponse,
    "PublishedFile.AddAppRelationship#1_Request": published_file.AddAppRelationshipRequest,
    "PublishedFile.AddAppRelationship#1_Response": published_file.AddAppRelationshipResponse,
    "PublishedFile.RemoveAppRelationship#1_Request": published_file.RemoveAppRelationshipRequest,
    "PublishedFile.RemoveAppRelationship#1_Response": published_file.RemoveAppRelationshipResponse,
    "PublishedFile.GetAppRelationships#1_Request": published_file.GetAppRelationshipsRequest,
    "PublishedFile.GetAppRelationships#1_Response": published_file.GetAppRelationshipsResponse,
    "PublishedFile.StartPlaytimeTracking#1_Request": published_file.StartPlaytimeTrackingRequest,
    "PublishedFile.StartPlaytimeTracking#1_Response": published_file.StartPlaytimeTrackingResponse,
    "PublishedFile.StopPlaytimeTracking#1_Request": published_file.StopPlaytimeTrackingRequest,
    "PublishedFile.StopPlaytimeTracking#1_Response": published_file.StopPlaytimeTrackingResponse,
    "PublishedFile.StopPlaytimeTrackingForAllAppItems#1_Request": published_file.StopPlaytimeTrackingForAllAppItemsRequest,
    "PublishedFile.StopPlaytimeTrackingForAllAppItems#1_Response": published_file.StopPlaytimeTrackingForAllAppItemsResponse,
    "PublishedFile.SetPlaytimeForControllerConfigs#1_Request": published_file.SetPlaytimeForControllerConfigsRequest,
    "PublishedFile.SetPlaytimeForControllerConfigs#1_Response": published_file.SetPlaytimeForControllerConfigsResponse,
    "PublishedFile.AddChild#1_Request": published_file.AddChildRequest,
    "PublishedFile.AddChild#1_Response": published_file.AddChildResponse,
    "PublishedFile.RemoveChild#1_Request": published_file.RemoveChildRequest,
    "PublishedFile.RemoveChild#1_Response": published_file.RemoveChildResponse,
    "PublishedFile.GetUserVoteSummary#1_Request": published_file.GetUserVoteSummaryRequest,
    "PublishedFile.GetUserVoteSummary#1_Response": published_file.GetUserVoteSummaryResponse,
    "PublishedFile.GetItemChanges#1_Request": published_file.GetItemChangesRequest,
    "PublishedFile.GetItemChanges#1_Response": published_file.GetItemChangesResponse,
    "TwoFactor.QueryStatus#1_Request": two_factor.StatusRequest,
    "TwoFactor.QueryStatus#1_Response": two_factor.StatusResponse,
    "TwoFactor.AddAuthenticator#1_Request": two_factor.AddAuthenticatorRequest,
    "TwoFactor.AddAuthenticator#1_Response": two_factor.AddAuthenticatorResponse,
    "TwoFactor.SendEmail#1_Request": two_factor.SendEmailRequest,
    "TwoFactor.SendEmail#1_Response": two_factor.SendEmailResponse,
    "TwoFactor.FinalizeAddAuthenticator#1_Request": two_factor.FinalizeAddAuthenticatorRequest,
    "TwoFactor.FinalizeAddAuthenticator#1_Response": two_factor.FinalizeAddAuthenticatorResponse,
    "TwoFactor.RemoveAuthenticator#1_Request": two_factor.RemoveAuthenticatorRequest,
    "TwoFactor.RemoveAuthenticator#1_Response": two_factor.RemoveAuthenticatorResponse,
    "TwoFactor.CreateEmergencyCodes#1_Request": two_factor.CreateEmergencyCodesRequest,
    "TwoFactor.CreateEmergencyCodes#1_Response": two_factor.CreateEmergencyCodesResponse,
    "TwoFactor.DestroyEmergencyCodes#1_Request": two_factor.DestroyEmergencyCodesRequest,
    "TwoFactor.DestroyEmergencyCodes#1_Response": two_factor.DestroyEmergencyCodesResponse,
    "TwoFactor.ValidateToken#1_Request": two_factor.ValidateTokenRequest,
    "TwoFactor.ValidateToken#1_Response": two_factor.ValidateTokenResponse,
    "UserReviews.Update#1_Request": reviews.UpdateRequest,
    "UserReviews.Update#1_Response": reviews.UpdateResponse,
    "UserReviews.GetIndividualRecommendations#1_Request": reviews.GetIndividualRecommendationsRequest,
    "UserReviews.GetIndividualRecommendations#1_Response": reviews.GetIndividualRecommendationsResponse,
}



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any

from .enums import UserBadge
from .game import StatefulGame

if TYPE_CHECKING:
    from .state import ConnectionState


__all__ = (
    "Badge",
    "FavouriteBadge",
    "UserBadges",
)


class BaseBadge:
    """
    Attributes
    ----------
    id
        The badge's ID.
    level
        The badge's level.
    game
        The game associated with the badge.
    """

    __slots__ = ()
    id: int
    level: int
    game: StatefulGame | None

    def __repr__(self) -> str:
        attrs = ("id", "level", "game")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.id == other.id


@dataclass(repr=False)
class FavouriteBadge(BaseBadge):
    """Represents a User's favourite badge.

    Attributes
    ----------
    id
        The badge's ID.
    level
        The badge's level.
    game
        The game associated with the badge.
    item_id
        The badge's item's ID.
    type
        The badge's type.
    border_colour
        The colour of the boarder of the badge.
    """

    __slots__ = ("id", "level", "game", "item_id", "type", "border_colour")

    id: int
    level: int
    game: StatefulGame | None
    item_id: int
    type: int
    border_colour: int


class Badge(BaseBadge):
    """Represents a Steam badge.

    Attributes
    ----------
    id
        The badge's ID.
    level
        The badge's level.
    game
        The game associated with the badge.
    xp
        The badge's XP.
    completion_time
        The time the badge was completed at.
    scarcity
        The scarcity of the badge.
    """

    __slots__ = ("id", "xp", "game", "level", "scarcity", "completion_time")

    def __init__(self, state: ConnectionState, data: dict[str, Any]):
        self.id = UserBadge.try_value(data["badgeid"])
        self.level: int = data["level"]
        self.xp: int = data["xp"]
        self.completion_time = datetime.utcfromtimestamp(data["completion_time"])
        self.scarcity: int = data["scarcity"]
        self.game = StatefulGame(state, id=data["appid"]) if "appid" in data else None


class UserBadges:
    """Represents a Steam :class:`~steam.User`'s badges/level.

    Attributes
    ----------
    level
        The badge's level.
    xp
        The badge's XP.
    xp_needed_to_level_up
        The amount of XP the user needs to level up.
    xp_needed_for_current_level
        The amount of XP the user's current level requires to achieve.
    badges
        A list of the user's badges.
    """

    __slots__ = (
        "xp",
        "level",
        "badges",
        "xp_needed_to_level_up",
        "xp_needed_for_current_level",
    )

    def __init__(self, state: ConnectionState, data: dict[str, Any]):
        self.level: int = data["player_level"]
        self.xp: int = data["player_xp"]
        self.xp_needed_to_level_up: int = data["player_xp_needed_to_level_up"]
        self.xp_needed_for_current_level: int = data["player_xp_needed_current_level"]
        self.badges: Sequence[Badge] = [Badge(state, data) for data in data["badges"]]

    def __repr__(self) -> str:
        attrs = ("level", "xp")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f'<UserBadges {" ".join(resolved)}>'

    def __len__(self) -> int:
        return len(self.badges)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .abc import SteamID
    from .clan import Clan
    from .enums import FriendRelationship
    from .state import ConnectionState
    from .user import User


__all__ = (
    "Invite",
    "UserInvite",
    "ClanInvite",
)


class Invite:
    __slots__ = ("invitee", "relationship", "_state")

    def __init__(self, state: ConnectionState, invitee: User | SteamID, relationship: FriendRelationship | None):
        self._state = state
        self.invitee = invitee
        self.relationship = relationship


class UserInvite(Invite):
    """Represents a invite from a user to become their friend.

    Attributes
    ----------
    invitee
        The user who sent the invite.
    relationship
        The relationship you have with the invitee. This ``None`` if the invite was sent while the bot is online.
    """

    __slots__ = ()

    def __repr__(self) -> str:
        attrs = ("invitee",)
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<UserInvite {' '.join(resolved)}>"

    async def accept(self) -> None:
        """Accept the invite request."""
        await self._state.http.accept_user_invite(self.invitee.id64)
        self._state.client.user.friends.append(self.invitee)

    async def decline(self) -> None:
        """Decline the invite request."""
        await self._state.http.decline_user_invite(self.invitee.id64)


class ClanInvite(Invite):
    """Represents a invite to join a :class:`~steam.Clan` from a user.

    Attributes
    ----------
    clan
        The clan to join.
    invitee
        The user who sent the invite.
    relationship
        The relationship you have with the clan.
    """

    __slots__ = ("clan",)

    def __init__(
        self,
        state: ConnectionState,
        invitee: User | SteamID,
        clan: Clan | SteamID,
        relationship: FriendRelationship | None,
    ):
        super().__init__(state, invitee, relationship)
        self.clan = clan

    def __repr__(self) -> str:
        attrs = ("invitee", "clan")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<ClanInvite {' '.join(resolved)}>"

    async def accept(self) -> None:
        """Accept the invite request."""
        await self._state.respond_to_clan_invite(self.clan.id64, True)

    async def decline(self) -> None:
        """Decline the invite request."""
        await self._state.respond_to_clan_invite(self.clan.id64, False)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import re
import time
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, overload

from bs4 import BeautifulSoup
from typing_extensions import Literal

from . import utils
from .abc import Commentable, SteamID
from .channel import ClanChannel
from .enums import ClanEvent, Type
from .errors import HTTPException, WSForbidden
from .event import Announcement, Event
from .game import Game, StatefulGame
from .iterators import AnnouncementsIterator, EventIterator
from .protobufs import chat
from .role import Role

if TYPE_CHECKING:
    from .state import ConnectionState
    from .user import User

__all__ = ("Clan",)


class Clan(SteamID, Commentable, utils.AsyncInit):
    """Represents a Steam clan.

    .. container:: operations

        .. describe:: x == y

            Checks if two clans are equal.

        .. describe:: str(x)

            Returns the clan's name.

    Attributes
    ------------
    name
        The name of the clan.
    chat_id
        The clan's chat id, this is different to :attr:`id`.
    avatar_url
        The icon url of the clan. Uses the large (184x184 px) image url.
    content
        The content of the clan.
    tagline
        The clan's tagline.
    member_count
        The amount of users in the clan.
    online_count
        The amount of users currently online.
    active_member_count
        The amount of currently users in the clan's chat room.
    in_game_count
        The amount of user's currently in game.
    created_at
        The time the clan was created_at.
    language
        The language set for the clan.
    location
        The location set for the clan.
    game
        The clan's associated game.
    owner
        The clan's owner.
    admins
        A list of the clan's administrators.
    mods
        A list of the clan's moderators.
    top_members
        A list of the clan's top_members.
    roles
        A list of the clan's roles.
    default_role
        The clan's default role.
    """

    __slots__ = (
        "name",
        "content",
        "avatar_url",
        "created_at",
        "language",
        "location",
        "member_count",
        "in_game_count",
        "online_count",
        "admins",
        "mods",
        "chat_id",
        "active_member_count",
        "owner",
        "default_role",
        "tagline",
        "top_members",
        "roles",
        "game",
        "community_url",
        "_channels",
        "_default_channel_id",
        "_state",
    )

    # TODO more to implement https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/groups.js
    # Clan.kick
    # Clan.ban
    # Group.create_channel
    # Clan.requesting_membership
    # Clan.respond_to_requesting_membership(*users, approve)
    # Clan.respond_to_all_requesting_membership(approve)

    name: str
    content: str
    avatar_url: str
    created_at: datetime | None
    member_count: int
    online_count: int
    in_game_count: int
    language: str
    location: str
    mods: list[User]
    admins: list[User]
    is_game_clan: bool
    community_url: str

    def __init__(self, state: ConnectionState, id: int):
        super().__init__(id, type=Type.Clan)
        self._state = state

        self.chat_id: int | None = None
        self.tagline: str | None = None
        self.game: StatefulGame | None = None
        self.owner: User | None = None
        self.active_member_count: int | None = None
        self.top_members: list[User | None] = []
        self.roles: list[Role] = []
        self.default_role: Role | None = None
        self._channels: dict[int, ClanChannel] = {}
        self._default_channel_id: int | None = None

    async def __ainit__(self) -> None:
        resp = await self._state.http._session.get(super().community_url)
        text = await resp.text()  # technically we loose proper request handling here
        if not self.id64:
            search = utils.CLAN_ID64_FROM_URL_REGEX.search(text)
            if search is None:
                raise ValueError("unreachable code reached")
            super().__init__(search["steamid"], type=Type.Clan)

        soup = BeautifulSoup(text, "html.parser")
        _, __, name = soup.title.text.rpartition(" :: ")
        self.name = name
        content = soup.find("meta", property="og:description")
        self.content = content["content"] if content is not None else None
        icon_url = soup.find("link", rel="image_src")
        self.avatar_url = icon_url["href"] if icon_url else None
        self.is_game_clan = "games" in resp.url.parts
        self.community_url = str(resp.url)
        if self.is_game_clan:
            for entry in soup.find_all("div", class_="actionItem"):
                a = entry.a
                if a is not None:
                    href = a.get("href", "")
                    match = re.findall(r"store.steampowered.com/app/(\d+)", href)
                    if match:
                        self.game = StatefulGame(self._state, id=match[0])
        stats = soup.find("div", class_="grouppage_resp_stats")
        if stats is None:
            return

        for stat in stats.find_all("div", class_="groupstat"):
            if "Founded" in stat.text:
                text = stat.text.split("Founded")[1].strip()
                if ", " not in stat.text:
                    text = f"{text}, {datetime.utcnow().year}"
                try:
                    self.created_at = datetime.strptime(text, "%d %B, %Y" if text.split()[0].isdigit() else "%B %d, %Y")
                except ValueError:  # why do other countries have to exist
                    self.created_at = None
            if "Language" in stat.text:
                self.language = stat.text.split("Language")[1].strip()
            if "Location" in stat.text:
                self.location = stat.text.split("Location")[1].strip()

        for count in stats.find_all("div", class_="membercount"):
            if "MEMBERS" in count.text:
                self.member_count = int(count.text.split("MEMBERS")[0].strip().replace(",", ""))
            if "IN-GAME" in count.text:
                self.in_game_count = int(count.text.split("IN-GAME")[0].strip().replace(",", ""))
            if "ONLINE" in count.text:
                self.online_count = int(count.text.split("ONLINE")[0].strip().replace(",", ""))

        admins = []
        mods = []
        is_admins = True
        for fields in soup.find_all("div", class_="membergrid"):
            for idx, field in enumerate(fields.find_all("div")):
                if "Members" in field.text:
                    if mods:
                        del mods[-1]
                    break
                if "Moderators" in field.text:
                    officer = admins.pop()
                    mods.append(officer)
                    is_admins = False
                try:
                    account_id = fields.find_all("div", class_="playerAvatar")[idx]["data-miniprofile"]
                except IndexError:
                    break
                else:
                    if is_admins:
                        admins.append(account_id)
                    else:
                        mods.append(account_id)

        users = await self._state.client.fetch_users(*admins, *mods)
        self.admins = [user for user in users if user and user.id in admins]
        self.admins = [user for user in users if user and user.id in mods]

    @classmethod
    async def _from_proto(
        cls,
        state: ConnectionState,
        clan_proto: chat.SummaryPair
        | chat.GetClanChatRoomInfoResponse
        | chat.GetChatRoomGroupSummaryResponse
        | chat.GroupHeaderState,
    ) -> Clan:
        if isinstance(clan_proto, chat.SummaryPair):
            id = clan_proto.group_summary.clanid
        else:
            id = clan_proto.chat_group_summary.clanid
        self = await cls(state, id)

        proto = clan_proto.group_summary if isinstance(clan_proto, chat.SummaryPair) else clan_proto.chat_group_summary

        self.chat_id = proto.chat_group_id
        self.tagline = proto.chat_group_tagline or None
        self.active_member_count = proto.active_member_count
        self.game = StatefulGame(self._state, id=proto.appid) if proto.appid else self.game

        self.owner = await self._state._maybe_user(utils.make_id64(proto.accountid_owner))
        self.top_members = await self._state.fetch_users([utils.make_id64(u) for u in proto.top_members])

        if hasattr(clan_proto, "roles"):
            roles = clan_proto.roles
        else:
            try:
                roles = await self._state.fetch_group_roles(self.chat_id)
            except WSForbidden:
                roles = []

        for role in roles:
            for permissions in proto.role_actions:
                if permissions.role_id == role.role_id:
                    self.roles.append(Role(self._state, self, role, permissions))
        self.default_role = utils.get(self.roles, id=proto.default_role_id)
        if not isinstance(clan_proto, chat.SummaryPair):
            return self

        for channel in clan_proto.user_chat_group_state.user_chat_room_state:
            try:
                new_channel = self._channels[channel.chat_id]
            except KeyError:
                new_channel = ClanChannel(state=self._state, clan=self, proto=channel)
                self._channels[new_channel.id] = new_channel
            else:
                new_channel._update(channel)
        self._default_channel_id = proto.default_chat_id
        return self

    def _update(self, proto: chat.ChatRoomGroupRoomsChangeNotification) -> None:
        for channel in proto.chat_rooms:
            try:
                new_channel = self._channels[channel.chat_id]
            except KeyError:
                new_channel = ClanChannel(state=self._state, clan=self, proto=channel)
                self._channels[new_channel.id] = new_channel
            else:
                new_channel._update(channel)
        self._default_channel_id = proto.default_chat_id

    def __repr__(self) -> str:
        attrs = ("name", "id", "chat_id", "type", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Clan {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    @property
    def _commentable_kwargs(self) -> dict[str, Any]:
        return {
            "id64": self.id64,
            "thread_type": 12,
        }

    @property
    def channels(self) -> list[ClanChannel]:
        """A list of the clan's channels."""
        return list(self._channels.values())

    @property
    def default_channel(self) -> ClanChannel | None:
        """The clan's default channel."""
        return self.get_channel(self._default_channel_id)  # type: ignore

    def get_channel(self, id: int) -> ClanChannel | None:
        """Get a channel from cache.

        Parameters
        ----------
        id
            The id of the channel.
        """
        return self._channels.get(id)

    async def fetch_members(self) -> list[SteamID]:
        """Fetches a clan's member list.

        Note
        ----
        This can be a very slow operation due to the rate limits on this endpoint.
        """

        def process(resp: str) -> BeautifulSoup:
            soup = BeautifulSoup(resp, "html.parser")
            for s in soup.find_all("div", id="memberList"):
                for user in s.find_all("div", class_="member_block"):
                    ret.append(SteamID(user["data-miniprofile"]))

            return soup

        async def getter(i: int) -> None:
            try:
                resp = await self._state.http.get(
                    f"{self.community_url}/members", params={"p": i + 1, "content_only": "true"}
                )
            except HTTPException:
                await asyncio.sleep(20)
                await getter(i)
            else:
                process(resp)

        ret = []
        resp = await self._state.http.get(f"{self.community_url}/members", params={"p": 1, "content_only": "true"})
        soup = process(resp)
        number_of_pages = int(re.findall(r"\d* - (\d*)", soup.find("div", class_="group_paging").text)[0])
        await asyncio.gather(*(getter(i) for i in range(1, number_of_pages)))
        return ret

    @property
    def description(self) -> str:
        """An alias to :attr:`content`.

        .. deprecated:: 0.8.0

            Use :attr:`content` instead.
        """
        warnings.warn("Clan.description is deprecated, use Clan.content instead", DeprecationWarning, stacklevel=2)
        return self.content

    @property
    def icon_url(self) -> str:
        """An alias to :attr:`avatar_url`.

        .. deprecated:: 0.8.0

            Use :attr:`avatar_url` instead.
        """
        warnings.warn("Clan.icon_url is deprecated, use Clan.avatar_url instead", DeprecationWarning, stacklevel=2)
        return self.avatar_url

    async def join(self) -> None:
        """Joins the clan. This will also join the clan's chat."""
        await self._state.http.join_clan(self.id64)
        await self._state.join_chat(self.chat_id)

    async def leave(self) -> None:
        """Leaves the clan."""
        await self._state.http.leave_clan(self.id64)

    async def invite(self, user: User) -> None:
        """Invites a :class:`~steam.User` to the clan.

        Parameters
        -----------
        user
            The user to invite to the clan.
        """
        await self._state.http.invite_user_to_clan(user.id64, self.id64)

    # event/announcement stuff

    async def fetch_event(self, id: int) -> Event:
        """Fetch an event from its ID.

        Parameters
        ----------
        id
            The ID of the event.
        """
        data = await self._state.http.get_clan_events(self.id, [id])
        events = data["events"]
        if not events:
            raise ValueError(f"Event {id} not found")
        return await Event(self._state, self, events[0])

    async def fetch_announcement(self, id: int) -> Announcement:
        """Fetch an announcement from its ID.

        Parameters
        ----------
        id
            The ID of the announcement.
        """
        data = await self._state.http.get_clan_announcement(self.id, id)
        announcement = data["events"][0]
        return await Announcement(self._state, self, announcement)

    def events(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> EventIterator:
        """An :class:`~steam.iterators.AsyncIterator` over a clan's :class:`steam.Event`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for event in client.events(limit=10):
                print(event.author, "made an event", event.name, "starting at", event.starts_at)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of events to search through. Default is ``100``. Setting this to ``None`` will fetch all
            of the clan's events, but this will be a very slow operation.
        before
            A time to search for events before.
        after
            A time to search for events after.

        Yields
        ---------
        :class:`~steam.Event`
        """
        return EventIterator(self, self._state, limit, before, after)

    def announcements(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AnnouncementsIterator:
        """An :class:`~steam.iterators.AsyncIterator` over a clan's :class:`steam.Announcement`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for announcement in client.announcements(limit=10):
                print(announcement.author, "made an announcement", announcement.name, "at", announcement.created_at)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of announcements to search through. Default is ``100``. Setting this to ``None`` will
            fetch all of the clan's announcements, but this will be a very slow operation.
        before
            A time to search for announcements before.
        after
            A time to search for announcements after.

        Yields
        ---------
        :class:`~steam.Announcement`
        """
        return AnnouncementsIterator(self, self._state, limit, before, after)

    @overload
    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: Literal[
            ClanEvent.Other,
            ClanEvent.Chat,
            ClanEvent.Party,
            ClanEvent.Meeting,
            ClanEvent.SpecialCause,
            ClanEvent.MusicAndArts,
            ClanEvent.Sports,
            ClanEvent.Trip,
        ] = ClanEvent.Other,
        starts_at: datetime | None = None,
    ) -> Event:
        ...

    @overload
    async def create_event(
        self,
        name: str,
        content: str,
        *,
        game: Game,
        type: Literal[ClanEvent.Game] = ...,
        starts_at: datetime | None = ...,
        server_address: str | None = ...,
        server_password: str | None = ...,
    ) -> Event:
        ...

    async def create_event(
        self,
        name: str,
        content: str,
        *,
        type: Literal[
            ClanEvent.Other,
            ClanEvent.Chat,
            ClanEvent.Game,
            # ClanEvent.Broadcast,  # TODO need to wait until implementing stream support for this
            ClanEvent.Party,
            ClanEvent.Meeting,
            ClanEvent.SpecialCause,
            ClanEvent.MusicAndArts,
            ClanEvent.Sports,
            ClanEvent.Trip,
        ] = ClanEvent.Other,
        game: Game | None = None,
        starts_at: datetime | None = None,
        server_address: str | None = None,
        server_password: str | None = None,
    ) -> Event:
        """Create an event.

        Parameters
        ----------
        name
            The name of the event
        content
            The content for the event.
        type
            The type of the event, defaults to :attr:`ClanEvent.Other`.
        game
            The game that will be played in the event. Required if type is :attr:`ClanEvent.Game`.
        starts_at
            The time the event will start at.
        server_address
            The address of the server that the event will be played on. This is only allowed if ``type`` is
            :attr:`ClanEvent.Game`.
        server_password
            The password for the server that the event will be played on. This is only allowed if ``type`` is
            :attr:`ClanEvent.Game`.

        Note
        ----
        It is recommended to use a timezone aware datetime for ``start``.

        Returns
        -------
        The created event.
        """

        resp = await self._state.http.create_clan_event(
            self.id64,
            name,
            content,
            f"{type.name}Event",
            str(game.id) if game is not None else "",
            server_address or "",
            server_password or "",
            starts_at,
        )
        soup = BeautifulSoup(resp, "html.parser")
        for element in soup.find_all("div", class_="eventBlockTitle"):
            a = element.a
            if a is not None and a.text == name:  # this is bad?
                _, __, id = a["href"].rpartition("/")
                if starts_at is not None:
                    timestamp = (
                        starts_at.timestamp()
                        if starts_at.tzinfo is not None
                        else (starts_at + (datetime.utcnow() - datetime.now())).timestamp()
                    )
                else:
                    timestamp = 0
                data = {
                    "gid": id,
                    "event_name": name,
                    "event_notes": content,
                    "event_type": type.value,
                    "appid": str(game.id) if game is not None else "",
                    "rtime32_start_time": timestamp,
                    "rtime32_last_modified": timestamp,
                    "server_address": server_address,
                    "server_password": server_password,
                }
                event = Event(self._state, self, data)
                event.author = self._state.client.user
                self._state.dispatch("event_create", event)
                return event
        raise ValueError

    async def create_announcement(
        self,
        name: str,
        content: str,
        hidden: bool = False,
    ) -> Announcement:
        """Create an announcement.

        Parameters
        ----------
        name
            The name of the announcement.
        content
            The content of the announcement.
        hidden
            Whether or not the announcement should initially be hidden.

        Returns
        -------
        The created announcement.
        """
        await self._state.http.create_clan_announcement(self.id64, name, content, hidden)
        resp = await self._state.http.get(f"{self.community_url}/announcements", params={"content_only": "true"})
        soup = BeautifulSoup(resp, "html.parser")
        for element in soup.find_all("div", class_="announcement"):
            a = element.a
            if a is not None and a.text == name:  # this is bad?
                _, __, id = a["href"].rpartition("/")
                timestamp = int(time.time())

                data = {
                    "gid": id,
                    "event_name": name,
                    "event_notes": "",
                    "event_type": ClanEvent.News,
                    "appid": 0,
                    "rtime32_start_time": timestamp,
                    "rtime32_last_modified": timestamp,
                    "announcement_body": {
                        "body": content,
                        "posttime": timestamp,
                        "updatetime": timestamp,
                    },
                    "comment_type": "ClanAnnouncement",
                    "hidden": hidden,
                }
                announcement = Announcement(self._state, self, data)
                announcement.author = self._state.client.user
                self._state.dispatch("announcement_create", announcement)
                return announcement

        raise ValueError



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a slightly modified version of discord.py's client
https://github.com/Rapptz/discord.py/blob/master/discord/client.py
"""

from __future__ import annotations

import asyncio
import datetime
import logging
import random
import sys
import time
import traceback
from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING, Any, TypeVar, overload

import aiohttp
from typing_extensions import Literal, ParamSpec, TypeAlias, final

from . import errors, utils
from .abc import SteamID
from .enums import Type
from .game import FetchedGame, Game, StatefulGame
from .game_server import GameServer, Query
from .gateway import *
from .guard import generate_one_time_code
from .http import HTTPClient
from .iterators import TradesIterator
from .models import TASK_HAS_NAME, URL, PriceOverview, return_true
from .state import ConnectionState
from .utils import make_id64

if TYPE_CHECKING:
    import steam

    from .abc import Message
    from .clan import Clan
    from .comment import Comment
    from .enums import PersonaState, PersonaStateFlag, UIMode
    from .event import Announcement, Event
    from .group import Group
    from .invite import ClanInvite, UserInvite
    from .protobufs import Msg, MsgProto
    from .trade import TradeOffer
    from .user import ClientUser, User

__all__ = ("Client",)

log = logging.getLogger(__name__)
EventType: TypeAlias = "Callable[..., Coroutine[Any, Any, Any]]"
EventDeco: TypeAlias = "Callable[[E], E] | E"
E = TypeVar("E", bound=EventType)
S = ParamSpec("S", bound="Client.start")  # 90% sure this is the way to use ParamSpec


class Client:
    """Represents a client connection that connects to Steam. This class is used to interact with the Steam API and CMs.

    Parameters
    ----------
    proxy
        A proxy URL to use for requests.
    proxy_auth
        The proxy authentication to use with requests.
    connector
        The connector to use with the :class:`aiohttp.ClientSession`.
    max_messages
        The maximum number of messages to store in the internal cache, default is 1000.
    game
        A games to set your status as on connect.
    games
        A list of games to set your status to on connect.
    state
        The state to show your account as on connect.

        Note
        ----
        Setting your status to :attr:`~steam.PersonaState.Offline`, will stop you receiving persona state updates
        and by extension :meth:`on_user_update` will stop being dispatched.
    ui_mode
        The UI mode to set your status to on connect.
    flags
        Flags to set your persona state to.
    force_kick
        Whether or not to forcefully kick any other playing sessions on connect. Defaults to ``False``.
    """

    # TODO
    # Client.create_clan
    # Client.create_group

    @overload
    def __init__(  # type: ignore
        self,
        *,
        proxy: str | None = ...,
        proxy_auth: aiohttp.BasicAuth | None = ...,
        connector: aiohttp.BaseConnector | None = ...,
        max_messages: int | None = ...,
        game: Game | None = ...,
        games: list[Game] = ...,
        state: PersonaState | None = ...,
        ui_mode: UIMode | None = ...,
        flags: PersonaStateFlag | None = ...,
        force_kick: bool = ...,
    ):
        ...

    def __init__(self, **options: Any):
        loop = options.get("loop")
        if loop:
            import inspect
            import warnings

            warnings.warn(
                "The loop argument is deprecated and scheduled for removal in V.1",
                stacklevel=len(inspect.stack())
                + 1,  # make sure its always at the top of the stack most likely where the Client was created
            )
        self.loop = asyncio.get_event_loop()
        self.http = HTTPClient(client=self, **options)
        self._connection = self._get_state(**options)
        self.ws: SteamWebSocket = None  # type: ignore

        self.username: str | None = None
        self.password: str | None = None
        self.shared_secret: str | None = None
        self.identity_secret: str | None = None

        self._closed = True
        self._cm_list: CMServerList | None = None
        self._listeners: dict[str, list[tuple[asyncio.Future, Callable[..., bool]]]] = {}
        self._ready = asyncio.Event()

    def _get_state(self, **options: Any) -> ConnectionState:
        return ConnectionState(client=self, **options)

    @property
    def user(self) -> ClientUser:
        """Represents the connected client. ``None`` if not logged in."""
        return self.http.user  # type: ignore

    @property
    def users(self) -> list[User]:
        """A list of all the users the connected client can see."""
        return self._connection.users

    @property
    def trades(self) -> list[TradeOffer]:
        """A list of all the trades the connected client can see."""
        return self._connection.trades

    @property
    def groups(self) -> list[Group]:
        """A list of all the groups the connected client is in."""
        return self._connection.groups

    @property
    def clans(self) -> list[Clan]:
        """A list of all the clans the connected client is in."""
        return self._connection.clans

    @property
    def latency(self) -> float:
        """Measures latency between a heartbeat send and the heartbeat interval in seconds."""
        return float("nan") if self.ws is None else self.ws.latency

    @property
    async def token(self) -> str:
        resp = await self.http.get(URL.COMMUNITY / "chat/clientjstoken")
        if not resp["logged_in"]:  # we got logged out :(
            await self.http.login(self.username, self.password, shared_secret=self.shared_secret)  # type: ignore
            return await self.token
        return resp["token"]

    async def code(self) -> str:
        """Get the current steam guard code.

        Warning
        -------
        This function will wait for a Steam guard code using :func:`input` in an executor if no shared_secret is passed
        to :meth:`run` or :meth:`start`, which blocks exiting until one is entered.
        """
        if self.shared_secret:
            return generate_one_time_code(self.shared_secret)
        print("Please enter a Steam guard code")
        code = await utils.ainput(">>> ")
        return code.strip()

    def is_ready(self) -> bool:
        """Specifies if the client's internal cache is ready for use."""
        return self._ready.is_set()

    def is_closed(self) -> bool:
        """Indicates if connection is closed to the API or CMs."""
        return self._closed

    @overload
    def event(self, coro: None = ...) -> Callable[[E], E]:
        ...

    @overload
    def event(self, coro: E) -> E:
        ...

    def event(self, coro: Callable[[E], E] | E | None = None) -> Callable[[E], E] | E:
        """|maybecallabledeco|
        Register an event to listen to.

        The events must be a :ref:`coroutine <coroutine>`, if not, :exc:`TypeError` is raised.

        Usage:

        .. code-block:: python3

            @client.event
            async def on_ready():
                print("Ready!")

        Raises
        ------
        :exc:`TypeError`
            The function passed is not a coroutine.
        """

        def decorator(coro: E) -> E:
            if not asyncio.iscoroutinefunction(coro):
                raise TypeError(f"Registered events must be a coroutines, {coro.__name__} is {type(coro).__name__}")

            setattr(self, coro.__name__, coro)
            log.debug(f"{coro.__name__} has been registered as an event")
            return coro

        return decorator(coro) if coro is not None else decorator

    async def _run_event(self, coro: EventType, event_name: str, *args: Any, **kwargs: Any) -> None:
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            try:
                await self.on_error(event_name, exc, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    def _schedule_event(self, coro: EventType, event_name: str, *args: Any, **kwargs: Any) -> asyncio.Task:
        wrapped = self._run_event(coro, event_name, *args, **kwargs)
        return (
            self.loop.create_task(wrapped, name=f"task_{event_name}")
            if TASK_HAS_NAME
            else self.loop.create_task(wrapped)
        )

    def dispatch(self, event: str, *args: Any, **kwargs: Any) -> None:
        log.debug(f"Dispatching event {event}")
        method = f"on_{event}"

        listeners = self._listeners.get(event)
        # remove the dispatched listener
        if listeners:
            removed = []
            for idx, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    removed.append(idx)
                    continue

                try:
                    result = condition(*args)
                except Exception as exc:
                    future.set_exception(exc)
                    removed.append(idx)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(idx)

            if len(removed) == len(listeners):
                self._listeners.pop(event)
            else:
                for idx in reversed(removed):
                    del listeners[idx]

        # schedule the event (if possible)
        try:
            coro = getattr(self, method)
        except AttributeError:
            pass
        else:
            self._schedule_event(coro, method, *args, **kwargs)

    async def _handle_ready(self) -> None:
        self._ready.set()
        self.dispatch("ready")

    @final
    def run(self, *args: S.args, **kwargs: S.kwargs) -> None:
        """A blocking call that abstracts away the event loop initialisation from you.

        It is not recommended to subclass this method, it is normally favourable to subclass :meth:`start` as it is a
        :ref:`coroutine <coroutine>`.

        Note
        ----
        This takes the same arguments as :meth:`start`.
        """

        async def runner() -> None:
            asyncio.events.new_event_loop = old_new_event_loop
            try:
                await self.start(*args, **kwargs)
            finally:
                if not self.is_closed():
                    await self.close()

        # we just have to monkey patch in support for using get_event_loop
        old_new_event_loop = asyncio.new_event_loop
        asyncio.events.new_event_loop = asyncio.get_event_loop
        try:
            asyncio.run(runner())
        except KeyboardInterrupt:
            log.info("Closing the event loop")

    async def login(self, username: str, password: str, *, shared_secret: str | None = None) -> None:
        """Login a Steam account and the Steam API with the specified credentials.

        Parameters
        ----------
        username
            The username of the user's account.
        password
            The password of the user's account.
        shared_secret
            The shared_secret of the desired Steam account, used to generate the 2FA code for login. If ``None`` is
            passed, the code will need to be inputted by the user via :meth:`code`.

        Raises
        ------
        :exc:`.InvalidCredentials`
            Invalid credentials were passed.
        :exc:`.LoginError`
            An unknown login related error occurred.
        :exc:`.NoCMsFound`
            No community managers could be found to connect to.
        """
        log.info("Logging in to steamcommunity.com")
        self.username = username
        self.password = password
        self.shared_secret = shared_secret

        await self.http.login(username, password, shared_secret=shared_secret)
        self._closed = False
        self.loop.create_task(self._connection.__ainit__())

    async def close(self) -> None:
        """Close the connection to Steam."""
        if self.is_closed():
            return

        self._closed = True

        if self.ws is not None:
            try:
                await self.change_presence(game=Game(id=0))  # disconnect from games
                await self.ws.handle_close()
            except ConnectionClosed:
                pass

        await self.http.close()
        self._ready.clear()

    def clear(self) -> None:
        """Clears the internal state of the bot. After this, the bot can be considered "re-opened", i.e.
        :meth:`is_closed` and :meth:`is_ready` both return ``False``. This also clears the internal cache.
        """
        self._closed = False
        self._ready.clear()
        self._connection.clear()
        self.http.clear()

    async def start(
        self,
        username: str,
        password: str,
        *,
        shared_secret: str | None = None,
        identity_secret: str | None = None,
    ) -> None:
        """A shorthand coroutine for :meth:`login` and :meth:`connect`.

        If no ``shared_secret`` is passed, you will have to manually enter a Steam guard code using :meth:`code`.

        Parameters
        ----------
        username
            The username of the account to login to.
        password
            The password of the account to login to.
        shared_secret
            The shared secret for the account to login to.
        identity_secret
            The identity secret for the account to login to.
        """

        self.username = username
        self.password = password
        self.shared_secret = shared_secret
        self.identity_secret = identity_secret

        if identity_secret is None:
            log.info("Trades will not be automatically accepted when sent as no identity_secret was passed.")

        await self.login(username, password, shared_secret=shared_secret)
        await self.connect()

    async def connect(self) -> None:
        """Initialize a connection to a Steam CM after logging in."""
        exceptions = (
            OSError,
            ConnectionClosed,
            aiohttp.ClientError,
            asyncio.TimeoutError,
            errors.HTTPException,
        )

        async def throttle() -> None:
            now = time.monotonic()
            between = now - last_connect
            sleep = random.random() * 4 if between > 600 else 100 / between ** 0.5
            log.info(f"Attempting to connect to another CM in {sleep}")
            await asyncio.sleep(sleep)

        while not self.is_closed():
            last_connect = time.monotonic()

            try:
                self.ws = await asyncio.wait_for(SteamWebSocket.from_client(self, cm_list=self._cm_list), timeout=60)
            except exceptions:
                await throttle()
                continue

            try:
                while True:
                    await self.ws.poll_event()
            except exceptions as exc:
                if isinstance(exc, ConnectionClosed):
                    self._cm_list = exc.cm_list
                self.dispatch("disconnect")
            finally:
                if not self.is_closed():
                    await throttle()

    # state stuff

    def get_user(self, id: utils.Intable) -> User | None:
        """Returns a user from cache with a matching ID or ``None`` if the user was not found.

        Parameters
        ----------
        id
            The ID of the user, can be an :attr:`.SteamID.id64`, :attr:`.SteamID.id`, :attr:`.SteamID.id2` or an
            :attr:`.SteamID.id3`.
        """
        id64 = make_id64(id=id, type=Type.Individual)
        return self._connection.get_user(id64)

    async def fetch_user(self, id: utils.Intable) -> User | None:
        """Fetches a user with a matching ID or ``None`` if the user was not found.

        Parameters
        ----------
        id
            The ID of the user, can be an :attr:`.SteamID.id64`, :attr:`.SteamID.id`, :attr:`.SteamID.id2` or an
            :attr:`.SteamID.id3`.
        """
        id64 = make_id64(id=id, type=Type.Individual)
        return await self._connection.fetch_user(id64)

    async def fetch_users(self, *ids: utils.Intable) -> list[User | None]:
        """Fetches a list of :class:`~steam.User` or ``None`` if the user was not found, from their IDs.

        Note
        ----
        The :class:`~steam.User` objects returned are unlikely to retain the order they were originally in.

        Parameters
        ----------
        ids
            The user's IDs.
        """
        id64s = [make_id64(id, type=Type.Individual) for id in ids]
        return await self._connection.fetch_users(id64s)

    async def fetch_user_named(self, name: str) -> User | None:
        """Fetches a user from https://steamcommunity.com from their community URL name.

        Parameters
        ----------
        name
            The name of the user after https://steamcommunity.com/id
        """
        id64 = await utils.id64_from_url(URL.COMMUNITY / f"id/{name}", self.http._session)
        return await self._connection.fetch_user(id64) if id64 is not None else None

    def get_trade(self, id: int) -> TradeOffer | None:
        """Get a trade from cache with a matching ID or ``None`` if the trade was not found.

        Parameters
        ----------
        id
            The id of the trade to search for from the cache.
        """
        return self._connection.get_trade(id)

    async def fetch_trade(self, id: int) -> TradeOffer | None:
        """Fetches a trade with a matching ID or ``None`` if the trade was not found.

        Parameters
        ----------
        id
            The ID of the trade to search for from the API.
        """
        return await self._connection.fetch_trade(id)

    def get_group(self, id: utils.Intable) -> Group | None:
        """Get a group from cache with a matching ID or ``None`` if the group was not found.

        Parameters
        ----------
        id
            The ID of the group, can be an :attr:`.SteamID.id64`, :attr:`.SteamID.id`, :attr:`.SteamID.id2` or an
            :attr:`.SteamID.id3`.
        """
        steam_id = SteamID(id=id, type=Type.Chat)
        return self._connection.get_group(steam_id.id)

    def get_clan(self, id: utils.Intable) -> Clan | None:
        """Get a clan from cache with a matching ID or ``None`` if the group was not found.

        Parameters
        ----------
        id
            The ID of the clan, can be an :attr:`.SteamID.id64`, :attr:`.SteamID.id`, :attr:`.SteamID.id2` or an
            :attr:`.SteamID.id3`.
        """
        steam_id = SteamID(id=id, type=Type.Clan)
        return self._connection.get_clan(steam_id.id)

    async def fetch_clan(self, id: utils.Intable) -> Clan | None:
        """Fetches a clan from the websocket with a matching ID or ``None`` if the clan was not found.

        Parameters
        ----------
        id
            The ID of the clan, can be an :attr:`.SteamID.id64`, :attr:`.SteamID.id`, :attr:`.SteamID.id2` or an
            :attr:`.SteamID.id3`.
        """
        id64 = make_id64(id=id, type=Type.Clan)
        return await self._connection.fetch_clan(id64)

    async def fetch_clan_named(self, name: str) -> Clan | None:
        """Fetches a clan from https://steamcommunity.com with a matching name or ``None`` if the clan was not found.

        Parameters
        ----------
        name
            The name of the Steam clan.
        """
        steam_id = await SteamID.from_url(URL.COMMUNITY / "clans" / name, self.http._session)
        return await self._connection.fetch_clan(steam_id.id64) if steam_id is not None else None

    def get_game(self, id: int | Game) -> StatefulGame:
        """Creates a stateful game from its ID.

        Parameters
        ----------
        id
            The app id of the game or a :class:`~steam.Game` instance.
        """
        return StatefulGame(self._connection, id=getattr(id, "id", id))

    async def fetch_game(self, id: int | Game) -> FetchedGame | None:
        """Fetch a game from its ID or ``None`` if the game was not found.

        Parameters
        ----------
        id
            The app id of the game or a :class:`~steam.Game` instance.
        """
        id = id if isinstance(id, int) else id.id
        resp = await self.http.get_game(id)
        if resp is None:
            return None
        data = resp[str(id)]
        if not data["success"]:
            return None
        return FetchedGame(self._connection, data["data"])

    @overload
    async def fetch_server(self, *, id: utils.Intable) -> GameServer | None:
        ...

    @overload
    async def fetch_server(
        self,
        *,
        ip: str,
        port: int | None = ...,
    ) -> GameServer | None:
        ...

    async def fetch_server(
        self,
        *,
        id: utils.Intable | None = None,
        ip: str | None = None,
        port: int | str | None = None,
    ) -> GameServer | None:
        """Fetch a :class:`.GameServer` from its ip and port or its SteamID or ``None`` if fetching the server failed.

        Parameters
        ----------
        ip
            The ip of the server.
        port
            The port of the server.
        id
            The ID of the game server, can be an :attr:`.SteamID.id64`, :attr:`.SteamID.id2` or an :attr:`.SteamID.id3`.
            If this is passed, it makes a call to the master server to fetch its ip and port.

        Note
        ----
        Passing an ``ip``, ``port`` and ``id`` to this function will raise an :exc:`TypeError`.
        """

        if all((id, ip, port)):
            raise TypeError("Too many arguments passed to fetch_server")
        if id:
            # we need to fetch the ip and port
            servers = await self._connection.fetch_server_ip_from_steam_id(make_id64(id, type=Type.GameServer))
            if not servers:
                raise ValueError(f"The master server didn't find a matching server for {id}")
            ip, _, port = servers[0].addr.partition(":")
        elif not ip:
            raise TypeError(f"fetch_server missing argument ip")

        servers = await self.fetch_servers(Query.ip / f"{ip}{f':{port}' if port is not None else ''}", limit=1)
        return servers[0] if servers else None

    async def fetch_servers(self, query: Query[Any], limit: int = 100) -> list[GameServer]:
        """Query game servers.

        Parameters
        ----------
        query
            The query to match servers with.
        limit
            The maximum amount of servers to return.
        """
        servers = await self._connection.fetch_servers(query.query, limit)
        return [GameServer(self._connection, server) for server in servers]

    # miscellaneous stuff

    def trade_history(
        self,
        limit: int | None = 100,
        before: datetime.datetime | None = None,
        after: datetime.datetime | None = None,
    ) -> TradesIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a :class:`steam.ClientUser`'s
        :class:`steam.TradeOffer` objects.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for trade in client.trade_history(limit=10):
                items = [getattr(item, "name", str(item.asset_id)) for item in trade.items_to_receive]
                items = ", ".join(items) or "Nothing"
                print("Partner:", trade.partner)
                print("Sent:", items)

        Flattening into a list:

        .. code-block:: python3

            trades = await client.trade_history(limit=50).flatten()
            # trades is now a list of TradeOffer

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of trades to search through. Default is ``100``. Setting this to ``None`` will fetch all
            of the user's trades, but this will be a very slow operation.
        before
            A time to search for trades before.
        after
            A time to search for trades after.

        Yields
        ---------
        :class:`~steam.TradeOffer`
        """
        return TradesIterator(state=self._connection, limit=limit, before=before, after=after)

    async def change_presence(
        self,
        *,
        game: Game | None = None,
        games: list[Game] | None = None,
        state: PersonaState | None = None,
        ui_mode: UIMode | None = None,
        flags: PersonaStateFlag | None = None,
        force_kick: bool = False,
    ) -> None:
        """Set your status.

        Parameters
        ----------
        game
            A games to set your status as.
        games
            A list of games to set your status to.
        state
            The state to show your account as.

            Warning
            -------
            Setting your status to :attr:`~steam.PersonaState.Offline`, will stop you receiving persona state updates
            and by extension :meth:`on_user_update` will stop being dispatched.

        ui_mode
            The UI mode to set your status to.
        flags
            The flags to update your account with.
        force_kick
            Whether or not to forcefully kick any other playing sessions.
        """
        games_ = [game.to_dict() for game in games] if games is not None else []
        if game is not None:
            games_.append(game.to_dict())
        await self.ws.change_presence(games=games_, state=state, flags=flags, ui_mode=ui_mode, force_kick=force_kick)

    async def trade_url(self, generate_new: bool = False) -> str:
        """Fetches this account's trade url.

        Parameters
        ----------
        generate_new
            Whether or not to generate a new trade token, defaults to ``False``.
        """
        return await self._connection.fetch_trade_url(generate_new)

    async def wait_until_ready(self) -> None:
        """Waits until the client's internal cache is all ready."""
        await self._ready.wait()

    async def fetch_price(self, name: str, game: Game, currency: int | None = None) -> PriceOverview:
        """Fetch the :class:`PriceOverview` for an item.

        Parameters
        ----------
        name
            The name of the item.
        game
            The game the item is from.
        currency
            The currency to fetch the price in.
        """
        price = await self.http.get_price(game.id, name, currency)
        return PriceOverview(price)

    # events to be subclassed

    async def on_error(self, event: str, error: Exception, *args, **kwargs):
        """The default error handler provided by the client.

        Usually when an event raises an uncaught exception, a traceback is printed to :attr:`sys.stderr` and the
        exception is ignored. If you want to change this behaviour and handle the exception yourself, this event can
        be overridden. Which, when done, will suppress the default action of printing the traceback.

        If you want exception to propagate out of the :class:`Client` class you can define an ``on_error`` handler
        consisting of a single empty :ref:`py:raise`. Exceptions raised by ``on_error`` will not be handled in any
        way by :class:`Client`.

        Parameters
        ----------
        event
            The name of the event that errored.
        error
            The error that was raised.
        args
            The positional arguments associated with the event.
        kwargs
            The key-word arguments associated with the event.
        """
        print(f"Ignoring exception in {event}", file=sys.stderr)
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    if TYPE_CHECKING or utils.DOCS_BUILDING:
        # these methods shouldn't exist at runtime unless subclassed to prevent pollution of logs

        async def on_connect(self) -> None:
            """Called when the client has successfully connected to Steam. This is not the same as the client being
            fully prepared, see :func:`on_ready` for that.

            The warnings on :meth:`on_ready` also apply.
            """

        async def on_disconnect(self) -> None:
            """Called when the client has disconnected from Steam. This could happen either through the internet
            disconnecting, an explicit call to logout, or Steam terminating the connection.

            This function can be called multiple times.
            """

        async def on_ready(self) -> None:
            """Called after a successful login and the client has handled setting up everything.

            Warning
            -------
            This function is not guaranteed to be the first event called. Likewise, this function is **not** guaranteed
            to only be called once. This library implements reconnection logic and will therefore end up calling this
            event whenever a CM disconnects.
            """

        async def on_login(self) -> None:
            """Called when the client has logged into https://steamcommunity.com."""

        async def on_logout(self) -> None:
            """Called when the client has logged out of https://steamcommunity.com."""

        async def on_message(self, message: "steam.Message") -> None:
            """Called when a message is created.

            Parameters
            ----------
            message
                The message that was received.
            """

        async def on_typing(self, user: "steam.User", when: "datetime.datetime") -> None:
            """Called when typing is started.

            Parameters
            ----------
            user
                The user that started typing.
            when
                The time the user started typing at.
            """

        async def on_trade_receive(self, trade: "steam.TradeOffer") -> None:
            """Called when the client receives a trade offer.

            Parameters
            ----------
            trade
                The trade offer that was received.
            """

        async def on_trade_send(self, trade: "steam.TradeOffer") -> None:
            """Called when the client sends a trade offer.

            Parameters
            ----------
            trade
                The trade offer that was sent.
            """

        async def on_trade_accept(self, trade: "steam.TradeOffer") -> None:
            """Called when the client or the trade partner accepts a trade offer.

            Parameters
            ----------
            trade
                The trade offer that was accepted.
            """

        async def on_trade_decline(self, trade: "steam.TradeOffer") -> None:
            """Called when the client or the trade partner declines a trade offer.

            Parameters
            ----------
            trade
                The trade offer that was declined.
            """

        async def on_trade_cancel(self, trade: "steam.TradeOffer") -> None:
            """Called when the client or the trade partner cancels a trade offer.

            Note
            ----
            This is called when the trade state becomes :attr:`~steam.TradeOfferState.Canceled` and
            :attr:`~steam.TradeOfferState.CanceledBySecondaryFactor`.

            Parameters
            ----------
            trade
                The trade offer that was cancelled.
            """

        async def on_trade_expire(self, trade: "steam.TradeOffer") -> None:
            """Called when a trade offer expires due to being active for too long.

            Parameters
            ----------
            trade
                The trade offer that expired.
            """

        async def on_trade_counter(self, trade: "steam.TradeOffer") -> None:
            """Called when the client or the trade partner counters a trade offer.

            Parameters
            ----------
            trade
                The trade offer that was countered.
            """

        async def on_comment(self, comment: "steam.Comment") -> None:
            """Called when the client receives a comment notification.

            Parameters
            ----------
            comment
                The comment received.
            """

        async def on_user_invite(self, invite: "steam.UserInvite") -> None:
            """Called when the client receives/sends an invite from/to a :class:`~steam.User` to become a friend.

            Parameters
            ----------
            invite
                The invite received.
            """

        async def on_user_invite_accept(self, invite: "steam.UserInvite") -> None:
            """Called when the client/invitee accepts an invite from/to a :class:`~steam.User` to become a friend.

            Parameters
            ----------
            invite
                The invite that was accepted.
            """

        async def on_user_invite_decline(self, invite: "steam.UserInvite") -> None:
            """Called when the client/invitee declines an invite from/to a :class:`~steam.User` to become a friend.

            Parameters
            ----------
            invite
                The invite that was declined.
            """

        async def on_user_update(self, before: "steam.User", after: "steam.User") -> None:
            """Called when a user is updated, due to one or more of the following attributes changing:

                - :attr:`~steam.User.name`
                - :attr:`~steam.User.state`
                - :attr:`~steam.User.flags`
                - :attr:`~steam.User.avatar_url`
                - :attr:`~steam.User.last_logon`
                - :attr:`~steam.User.last_logoff`
                - :attr:`~steam.User.last_seen_online`
                - :attr:`~steam.User.game`

            Parameters
            ----------
            before
                The user's state before it was updated.
            after
                The user's state now.
            """

        async def on_user_remove(self, user: "steam.User") -> None:
            """Called when you or the ``user`` remove each other from your friends lists.

            Parameters
            ----------
            user
                The user who was removed.
            """

        async def on_clan_invite(self, invite: "steam.ClanInvite") -> None:
            """Called when the client receives/sends an invite from/to a :class:`~steam.User` to join a
            :class:`~steam.Clan`.

            Parameters
            ----------
            invite
                The invite received.
            """

        async def on_clan_invite_accept(self, invite: "steam.ClanInvite") -> None:
            """Called when the client/invitee accepts an invite to join a :class:`~steam.Clan`.

            Parameters
            ----------
            invite
                The invite that was accepted.
            """

        async def on_clan_invite_decline(self, invite: "steam.ClanInvite") -> None:
            """Called when the client/invitee declines an invite to join a :class:`~steam.Clan`.

            Parameters
            ----------
            invite
                The invite that was declined.
            """

        async def on_clan_join(self, clan: "steam.Clan") -> None:
            """Called when the client joins a new clan.

            Parameters
            ----------
            clan
                The joined clan.
            """

        async def on_clan_update(self, before: "steam.Clan", after: "steam.Clan") -> None:
            """Called when a clan is updated, due to one or more of the following attributes changing:

                - :attr:`~steam.Clan.name`
                - :attr:`~steam.Clan.avatar_url`
                - :attr:`~steam.Clan.member_count`
                - :attr:`~steam.Clan.in_game_count`
                - :attr:`~steam.Clan.online_count`
                - :attr:`~steam.Clan.active_member_count`

            Parameters
            ----------
            before
                The clan's state before it was updated.
            after
                The clan's state now.
            """

        async def on_clan_leave(self, clan: "steam.Clan") -> None:
            """Called when the client leaves a clan.

            Parameters
            ----------
            clan
                The left clan.
            """

        async def on_group_join(self, group: "steam.Group") -> None:
            """Called when the client joins a new group.

            Parameters
            ----------
            group
                The joined group.
            """

        async def on_group_update(self, before: "steam.Group", after: "steam.Group") -> None:
            """Called when a group is updated.

            Parameters
            ----------
            before
                The group's state before it was updated.
            after
                The group's state now.
            """

        async def on_group_leave(self, group: "steam.Group") -> None:
            """Called when the client leaves a group.

            Parameters
            ----------
            group
                The left group.
            """

        async def on_event_create(self, event: "steam.Event") -> None:
            """Called when an event in a clan is created.

            Parameters
            ----------
            event
                The event that was created.
            """

        async def on_announcement_create(self, announcement: "steam.Announcement") -> None:
            """Called when an announcement in a clan is created.

            Parameters
            ----------
            announcement
                The announcement that was created.
            """

        async def on_socket_receive(self, msg: "Msg | MsgProto") -> None:
            """Called when the connected CM parses a received ``Msg``/``MsgProto``

            Parameters
            ----------
            msg
                The received message.
            """

        async def on_socket_send(self, msg: "Msg | MsgProto") -> None:
            """Called when the client sends a ``Msg``/``MsgProto`` to the connected CM.

            Parameters
            ----------
            msg
                The sent message.
            """

    @overload
    async def wait_for(
        self,
        event: Literal[
            "connect",
            "disconnect",
            "ready",
            "login",
            "logout",
        ],
        *,
        check: Callable[[], bool] = ...,
        timeout: float | None = ...,
    ) -> None:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["error"],
        *,
        check: Callable[[str, Exception, tuple[Any, ...], dict[str, Any]], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[str, Exception, tuple[Any, ...], dict[str, Any]]:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["message"],
        *,
        check: Callable[[Message], bool] = ...,
        timeout: float | None = ...,
    ) -> Message:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["comment"],
        *,
        check: Callable[[Comment], bool] = ...,
        timeout: float | None = ...,
    ) -> Comment:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["user_update"],
        *,
        check: Callable[[User, User], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[User, User]:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["clan_update"],
        *,
        check: Callable[[Clan, Clan], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[Clan, Clan]:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["group_update"],
        *,
        check: Callable[[Group, Group], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[Group, Group]:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["typing"],
        *,
        check: Callable[[User, datetime.datetime], bool] = ...,
        timeout: float | None = ...,
    ) -> tuple[User, datetime.datetime]:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "trade_receive",
            "trade_send",
            "trade_accept",
            "trade_decline",
            "trade_cancel",
            "trade_expire",
            "trade_counter",
        ],
        *,
        check: Callable[[TradeOffer], bool] = ...,
        timeout: float | None = ...,
    ) -> TradeOffer:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "user_invite",
            "user_invite_accept",
            "user_invite_decline",
        ],
        *,
        check: Callable[[UserInvite], bool] = ...,
        timeout: float | None = ...,
    ) -> UserInvite:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["user_remove"],
        *,
        check: Callable[[User], bool] = ...,
        timeout: float | None = ...,
    ) -> User:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "clan_invite",
            "clan_invite_accept",
            "clan_invite_decline",
        ],
        *,
        check: Callable[[ClanInvite], bool] = ...,
        timeout: float | None = ...,
    ) -> ClanInvite:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "clan_join",
            "clan_leave",
        ],
        *,
        check: Callable[[Clan], bool] = ...,
        timeout: float | None = ...,
    ) -> Clan:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "group_join",
            "group_leave",
        ],
        *,
        check: Callable[[Group], bool] = ...,
        timeout: float | None = ...,
    ) -> Group:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["event_create"],
        *,
        check: Callable[[Event], bool] = ...,
        timeout: float | None = ...,
    ) -> Event:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal["announcement_create"],
        *,
        check: Callable[[Announcement], bool] = ...,
        timeout: float | None = ...,
    ) -> Announcement:
        ...

    @overload
    async def wait_for(
        self,
        event: Literal[
            "socket_receive",
            "socket_send",
        ],
        *,
        check: Callable[[Msgs], bool] = ...,
        timeout: float | None = ...,
    ) -> Msgs:
        ...

    async def wait_for(
        self,
        event: str,
        *,
        check: Callable[..., bool] = return_true,
        timeout: float | None = None,
    ) -> Any:
        """Wait for the first event to be dispatched that meets the requirements, this by default is the first event
        with a matching event name.

        Parameters
        ----------
        event
            The event name from the :ref:`event reference <event-reference>`, but without the ``on_`` prefix, to wait
            for.
        check
            A callable predicate that checks the received event. The arguments must match the parameters of the
            ``event`` being waited for and must return a :class:`bool`.
        timeout
            By default, :meth:`wait_for` function does not timeout, however, in the case a ``timeout`` parameter is
            passed after the amount of seconds pass :exc:`asyncio.TimeoutError` is raised.

        Raises
        ------
        :exc:`asyncio.TimeoutError`
            If the provided timeout was reached.

        Returns
        -------
        Returns ``None``, a single argument or a :class:`tuple` of multiple arguments that mirrors the parameters for
        the ``event`` parameter from the :ref:`event reference <event-reference>`.
        """
        future = self.loop.create_future()

        event_lower = event.lower()
        try:
            listeners = self._listeners[event_lower]
        except KeyError:
            listeners = []
            self._listeners[event_lower] = listeners

        listeners.append((future, check))
        return await asyncio.wait_for(future, timeout)



"""
The MIT License (MIT)

Copyright (c) 2015 Rossen Georgiev <rossen@rgp.io>
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is an updated version of https://github.com/ValvePython/steam/tree/master/steam/core/msg
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import betterproto
from typing_extensions import TypeAlias

from ..enums import IntEnum, Result
from .emsg import *
from .headers import *
from .protobufs import *
from .unified import *

M = TypeVar("M", bound=betterproto.Message, covariant=True)
GetProtoType: TypeAlias = "type[betterproto.Message] | None"
ALLOWED_HEADERS = (ExtendedMsgHdr, MsgHdrProto, GCMsgHdrProto)
betterproto.safe_snake_case = do_nothing_case


def get_cmsg(msg: IntEnum) -> GetProtoType:
    """Get a protobuf from its EMsg."""
    try:
        return PROTOBUFS[msg]
    except KeyError:
        return None


def get_um(name: str, request: bool = True) -> GetProtoType:
    """Get the protobuf for a certain Unified Message."""
    try:
        return UMS[f"{name}#1_{'Request' if request else 'Response'}"]
    except KeyError:
        return None


class MsgBase(Generic[M]):
    __slots__ = ("header", "body", "payload", "skip")

    if sys.version_info < (3, 9):  # see https://bugs.python.org/issue39168 for the rational behind this

        def __new__(cls, *args: Any, **kwargs: Any):  # type: ignore
            return object.__new__(cls)

    def __init__(self, msg: IntEnum, data: bytes | None, **kwargs: Any):
        self.msg = msg
        self.body: M = None  # type: ignore
        self.payload: bytes | None = data[self.skip :] if data else None

        self.parse()
        if kwargs and self.body is not None:
            self.body.from_dict(kwargs)

    def __repr__(self) -> str:
        attrs = ("msg", "header")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        if self.body is not None:
            resolved.extend(f"{k}={v!r}" for k, v in self.body.to_dict(do_nothing_case).items())
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __bytes__(self) -> bytes:
        return bytes(self.header) + bytes(self.body)

    def _parse(self, proto: type[M] | None) -> None:
        if proto:
            self.body = proto()
            if self.payload:
                self.body.parse(self.payload)

    @property
    def msg(self) -> IntEnum:
        return self.header.msg

    @msg.setter
    def msg(self, value: IntEnum) -> None:
        self.header.msg = value

    @property
    def steam_id(self) -> int:
        return self.header.body.steam_id

    @steam_id.setter
    def steam_id(self, value: int) -> None:
        self.header.body.steam_id = value

    @property
    def session_id(self) -> int:
        return self.header.body.session_id

    @session_id.setter
    def session_id(self, value: int) -> None:
        self.header.body.session_id = value

    @property
    def result(self) -> Result:
        return Result.try_value(getattr(self.body, "eresult", 0) or self.header.body.eresult)

    def parse(self) -> None:
        proto = get_cmsg(self.msg)
        self._parse(proto)


if not TYPE_CHECKING:
    MsgBase.__class_getitem__ = classmethod(
        lambda cls, params: cls
    )  # really don't want this to be picked up by linters


class Msg(MsgBase[M]):
    """A wrapper around received messages."""

    __slots__ = ()

    def __init__(
        self,
        msg: EMsg,
        data: bytes | None = None,
        extended: bool = False,
        **kwargs: Any,
    ):
        self.header = ExtendedMsgHdr(data) if extended else MsgHdr(data)
        self.skip = self.header.STRUCT.size
        super().__init__(msg, data, **kwargs)


class MsgProto(MsgBase[M]):
    """A wrapper around received protobuf messages."""

    __slots__ = ()

    def __init__(
        self,
        msg: EMsg,
        data: bytes | None = None,
        um_name: str | None = None,
        **kwargs: Any,
    ):
        self.header = MsgHdrProto(data)
        if um_name:
            self.header.body.job_name_target = um_name
        self.skip = self.header.length
        super().__init__(msg, data, **kwargs)

    def parse(self) -> None:
        if self.msg in (
            EMsg.ServiceMethod,
            EMsg.ServiceMethodResponse,
            EMsg.ServiceMethodSendToClient,
            EMsg.ServiceMethodCallFromClient,
        ):
            name = self.header.body.job_name_target
            if name:
                name, _, __ = name.partition("#")
                self.header.body.job_name_target = name
                proto = get_um(name, self.msg in (EMsg.ServiceMethod, EMsg.ServiceMethodCallFromClient))
            else:
                proto = None

        else:
            proto = get_cmsg(self.msg)

        self._parse(proto)


class GCMsg(MsgBase[M]):
    """A wrapper around received GC messages, mainly for extensions."""

    __slots__ = ()

    def __init__(
        self,
        msg: IntEnum,
        data: bytes | None = None,
        **kwargs: Any,
    ):
        self.header = GCMsgHdr(data)
        self.skip = self.header.STRUCT.size
        super().__init__(msg, data, **kwargs)


class GCMsgProto(MsgBase[M]):
    """A wrapper around received GC protobuf messages, mainly for extensions."""

    __slots__ = ()

    def __init__(
        self,
        msg: IntEnum,
        data: bytes | None = None,
        **kwargs: Any,
    ):
        self.header = GCMsgHdrProto(data)
        self.skip = self.header.length
        super().__init__(msg, data, **kwargs)



"""
steam.py
~~~~~~~~~

A basic wrapper for the Steam API and its Community Managers.
"""

from . import abc as abc, guard as guard, utils as utils
from .__metadata__ import *
from .abc import *
from .badge import *
from .channel import *
from .clan import *
from .client import *
from .comment import *
from .enums import *
from .errors import *
from .event import *
from .game import *
from .game_server import *
from .group import *
from .image import *
from .invite import *
from .message import *
from .models import *
from .profile import *
from .role import *
from .trade import *
from .user import *

__import__("logging").getLogger(__name__).addHandler(__import__("logging").NullHandler())  # don't leak scope
__import__("warnings").filterwarnings("always", ".*", module=rf"{__name__}(.\w+)+", append=False)

__getattr__ = enums.__getattr__  # shim for old enum names



# These were removed for some reason, pulled from:
# https://github.com/SteamDatabase/Protobufs/blob/master/webui/service_community.proto and its history.
# Last updated 9/9/2021

from dataclasses import dataclass
from typing import List

import betterproto

from .base import CMsgIpAddress


@dataclass(eq=False, repr=False)
class GetCommentThreadRequest(betterproto.Message):
    id64: int = betterproto.fixed64_field(1)
    thread_type: int = betterproto.uint32_field(2)
    gidfeature: int = betterproto.fixed64_field(3)
    gidfeature2: int = betterproto.fixed64_field(4)
    thread_id: int = betterproto.fixed64_field(5)
    start: int = betterproto.int32_field(6)
    count: int = betterproto.int32_field(7)
    upvotes: int = betterproto.int32_field(8)
    include_deleted: bool = betterproto.bool_field(9)
    id: int = betterproto.fixed64_field(10)
    time_oldest: int = betterproto.uint32_field(11)
    oldest_first: bool = betterproto.bool_field(12)


@dataclass(eq=False, repr=False)
class GetCommentThreadResponse(betterproto.Message):
    @dataclass(eq=False, repr=False)
    class Comment(betterproto.Message):
        id: int = betterproto.fixed64_field(1)
        author_id64: int = betterproto.fixed64_field(2)
        timestamp: int = betterproto.uint32_field(3)
        content: str = betterproto.string_field(4)
        upvotes: int = betterproto.int32_field(5)
        hidden: bool = betterproto.bool_field(6)
        hidden_by_user: bool = betterproto.bool_field(7)
        deleted: bool = betterproto.bool_field(8)
        ipaddress: CMsgIpAddress = betterproto.message_field(9)
        total_hidden: int = betterproto.int32_field(10)
        upvoted_by_user: bool = betterproto.bool_field(11)

        @dataclass(eq=False, repr=False)
        class Reaction(betterproto.Message):
            id: int = betterproto.uint32_field(1)
            count: int = betterproto.uint32_field(1)

        reactions: List[Reaction] = betterproto.message_field(12)
        parent_id: int = betterproto.fixed64_field(13)

    comments: List[Comment] = betterproto.message_field(1)
    deleted_comments: List[Comment] = betterproto.message_field(2)
    id64: int = betterproto.fixed64_field(3)
    thread_id: int = betterproto.fixed64_field(4)
    start: int = betterproto.int32_field(5)
    count: int = betterproto.int32_field(6)
    total_count: int = betterproto.int32_field(7)
    upvotes: int = betterproto.int32_field(8)
    upvoters: List[int] = betterproto.uint32_field(9)
    user_subscribed: bool = betterproto.bool_field(10)
    user_upvoted: bool = betterproto.bool_field(11)
    answer_id: int = betterproto.fixed64_field(12)
    answer_actor: int = betterproto.uint32_field(13)
    answer_actor_rank: int = betterproto.int32_field(14)
    can_post: bool = betterproto.bool_field(15)
    thread_type: int = betterproto.uint32_field(16)
    gidfeature: int = betterproto.fixed64_field(17)
    gidfeature2: int = betterproto.fixed64_field(18)


@dataclass(eq=False, repr=False)
class PostCommentToThreadRequest(betterproto.Message):
    id64: int = betterproto.fixed64_field(1)
    thread_type: int = betterproto.uint32_field(2)
    gidfeature: int = betterproto.fixed64_field(3)
    gidfeature2: int = betterproto.fixed64_field(4)
    content: str = betterproto.string_field(6)
    parent_id: int = betterproto.fixed64_field(7)
    suppress_notifications: bool = betterproto.bool_field(8)
    is_report: bool = betterproto.bool_field(9)


@dataclass(eq=False, repr=False)
class PostCommentToThreadResponse(betterproto.Message):
    id: int = betterproto.fixed64_field(1)
    thread_id: int = betterproto.fixed64_field(2)
    count: int = betterproto.int32_field(3)
    upvotes: int = betterproto.int32_field(4)


@dataclass(eq=False, repr=False)
class DeleteCommentFromThreadRequest(betterproto.Message):
    id64: int = betterproto.fixed64_field(1)
    thread_type: int = betterproto.uint32_field(2)
    gidfeature: int = betterproto.fixed64_field(3)
    gidfeature2: int = betterproto.fixed64_field(4)
    id: int = betterproto.fixed64_field(5)
    undelete: bool = betterproto.bool_field(6)


@dataclass(eq=False, repr=False)
class DeleteCommentFromThreadResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RateCommentThreadRequest(betterproto.Message):
    thread_type: str = betterproto.string_field(1)
    id64: int = betterproto.fixed64_field(2)
    gidfeature: int = betterproto.fixed64_field(3)
    gidfeature2: int = betterproto.fixed64_field(4)
    id: int = betterproto.fixed64_field(5)
    rate_up: bool = betterproto.bool_field(6)
    suppress_notifications: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class RateCommentThreadResponse(betterproto.Message):
    id: int = betterproto.fixed64_field(1)
    thread_id: int = betterproto.uint64_field(2)
    count: int = betterproto.uint32_field(3)
    upvotes: int = betterproto.uint32_field(4)
    has_upvoted: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class GetCommentThreadRatingsRequest(betterproto.Message):
    thread_type: str = betterproto.string_field(1)
    id64: int = betterproto.uint64_field(2)
    gidfeature: int = betterproto.uint64_field(3)
    gidfeature2: int = betterproto.uint64_field(4)
    comment_id: int = betterproto.uint64_field(5)
    max_results: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class GetCommentThreadRatingsResponse(betterproto.Message):
    thread_id: int = betterproto.uint64_field(1)
    comment_id: int = betterproto.uint64_field(2)
    upvotes: int = betterproto.uint32_field(3)
    has_upvoted: bool = betterproto.bool_field(4)
    upvoter_ids: List[int] = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class RateClanAnnouncementRequest(betterproto.Message):
    announcementid: int = betterproto.uint64_field(1)
    vote_up: bool = betterproto.bool_field(2)
    clan_accountid: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class RateClanAnnouncementResponse(betterproto.Message):
    pass



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import re
import warnings
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, TypeVar, overload

from typing_extensions import Literal, TypedDict

from . import utils
from .enums import Enum_, ReviewType
from .models import URL
from .utils import Intable, id64_from_url

if TYPE_CHECKING:
    from .clan import Clan
    from .state import ConnectionState
    from .user import User

__all__ = (
    "TF2",
    "Game",
    "CSGO",
    "DOTA2",
    "LFD2",
    "STEAM",
    "CUSTOM_GAME",
    "UserGame",
    "WishlistGame",
    "FetchedGame",
)

T = TypeVar("T")
APP_ID_MAX = 2 ** 32


class GameDict(TypedDict):
    name: str
    appid: str
    playtime_forever: int
    img_icon_url: str
    img_logo_url: str
    has_community_visible_stats: bool


class GameToDict(TypedDict, total=False):
    game_id: str
    game_extra_info: str


class WishlistGameDict(TypedDict):
    name: str
    capsule: str
    review_score: int
    review_desc: str
    reviews_total: str
    reviews_percent: int
    release_date: int
    release_string: str
    platform_icons: str
    subs: list[dict[str, Any]]
    type: str
    screenshots: list[str]
    review_css: str
    priority: int
    added: int
    background: str
    rank: int
    tags: list[str]
    is_free_game: bool
    win: bool
    mac: bool
    linux: bool


class PackageGroupsDict(TypedDict):
    name: Literal["default", "subscriptions"]
    title: str
    description: str
    selection_text: str
    display_type: Literal[0, 1]
    is_recurring_subscription: bool
    subs: list[dict[str, str]]


class FetchedGameDict(TypedDict):
    # https://wiki.teamfortress.com/wiki/User:RJackson/StorefrontAPI#Result_data_3
    type: Literal["game", "dlc", "demo", "advertising", "mod", "video"]
    name: str
    steam_appid: int
    required_age: int
    is_free: bool
    controller_support: Literal["partial", "full"]
    dlc: list[int]
    detailed_description: str
    short_description: str
    fullgame: dict[str, Any]
    supported_languages: str
    header_image: str
    pc_requirements: list[dict[str, str]]
    mac_requirements: list[dict[str, str]]
    linux_requirements: list[dict[str, str]]
    legal_notice: str
    developers: list[str]
    publishers: list[str]
    demos: list[dict[str, Any]]
    price_overview: list[dict[str, Any]]
    package_groups: list[PackageGroupsDict]
    platforms: dict[str, bool]
    metacritic: list[dict[str, str]]
    categories: list[dict[str, str]]
    release_date: dict[str, str]
    background: str
    website: str
    movies: list[dict[str, Any]]


class Game:
    """Represents a Steam game.

    Attributes
    ----------
    name
        The game's name.
    id
        The game's app ID.
    context_id
        The context id of the game normally ``2``.
    """

    __slots__ = (
        "id",
        "name",
        "context_id",
    )

    @overload
    def __init__(self, *, id: Intable, name: str | None = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(self, *, name: Literal["Team Fortress 2"], id: Intable = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(self, *, name: Literal["Left 4 Dead 2"], id: Intable = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(self, *, name: Literal["DOTA 2"], id: Intable = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(
        self, *, name: Literal["Counter Strike Global-Offensive"], id: Intable = ..., context_id: int | None = ...
    ):
        ...

    @overload
    def __init__(self, *, name: Literal["Steam"], id: Intable = ..., context_id: int | None = ...):
        ...

    def __init__(
        self,
        *,
        id: Intable | None = None,
        name: str | None = None,
        title: str | None = None,
        context_id: int | None = None,
    ):
        if title is not None:
            warnings.warn("Game.title is depreciated, use Game.name instead", DeprecationWarning)
        name = name or title
        if name is None and id is None:
            raise TypeError("__init__() missing a required keyword argument: 'id' or 'name'")

        if name is None:
            try:
                id = int(id)
            except (ValueError, TypeError):
                raise ValueError("id expected to support int()")
            try:
                name = Games(id).name
            except ValueError:
                name = None
            else:
                if name == "Steam" and context_id is None:
                    context_id = 6
        elif id is None:
            id = utils.get(Games, name=name)
            if id is None:
                raise ValueError(f"Cannot find a matching game for {name!r}")
        else:
            try:
                id = int(id)
            except (ValueError, TypeError):
                raise ValueError("id must be an int") from None

        if id < 0:
            raise ValueError("id cannot be negative")

        self.id: int = id
        self.name: str | None = name
        self.context_id: int = 2 if context_id is None else context_id

    def __str__(self) -> str:
        return self.name or ""

    def __repr__(self) -> str:
        attrs = ("name", "id", "context_id")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"{self.__class__.__name__}({', '.join(resolved)})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Game):
            return NotImplemented
        return self.id == other.id

    @property
    def title(self) -> str | None:
        """The game's name.

        .. deprecated:: 0.8.0

            Use :attr:`name` instead.
        """
        warnings.warn("Game.title is depreciated, use Game.name instead", DeprecationWarning, stacklevel=2)
        return self.name

    def to_dict(self) -> GameToDict:
        if self.is_steam_game():
            return {"game_id": str(self.id)}

        if self.name is None or self.id is None:
            raise TypeError("unserializable game with no title and or id")
        return {"game_id": str(self.id), "game_extra_info": self.name}

    def is_steam_game(self) -> bool:
        """Whether the game could be a Steam game."""
        return bool(self.id is not None and self.id <= APP_ID_MAX)

    @property
    def url(self) -> str:
        """What should be the game's url on steamcommunity if applicable."""
        return f"{URL.COMMUNITY}/app/{self.id}"


class Games(Game, Enum_):
    """This is "enum" to trick type checkers into allowing Literal[TF2] to be valid for overloads in extensions."""

    __slots__ = ("_name",)

    def __new__(cls, *, name: str, value: tuple[str, int, int]) -> Games:
        self = object.__new__(cls)
        set_attribute = object.__setattr__
        set_attribute(self, "_name", name)
        set_attribute(self, "name", value[0])
        set_attribute(self, "id", value[1])
        set_attribute(self, "context_id", value[2])
        return self

    def __repr__(self) -> str:
        return self._name

    TF2 = "Team Fortress 2", 440, 2
    LFD2 = "Left 4 Dead 2", 550, 2
    DOTA2 = "DOTA 2", 570, 2
    CSGO = "Counter Strike Global-Offensive", 730, 2
    STEAM = "Steam", 753, 6


TF2 = Games.TF2
DOTA2 = Games.DOTA2
CSGO = Games.CSGO
LFD2 = Games.LFD2
STEAM = Games.STEAM


@overload
def CUSTOM_GAME(name: str) -> Game:  # type: ignore
    ...


def CUSTOM_GAME(name: str = None, title: str = None) -> Game:
    """Create a custom game instance for :meth:`~steam.Client.change_presence`.
    The :attr:`Game.id` will be set to ``15190414816125648896`` and the :attr:`Game.context_id` to ``None``.

    Example:

    .. code-block:: python3

        await client.change_presence(game=steam.CUSTOM_GAME("my cool game"))

    Parameters
    ----------
    name
        The name of the game to set your playing status to.
    """
    if name is None and title is None:
        raise TypeError("CUSTOM_GAME missing argument 'name'")
    return Game(name=name or title, id=15190414816125648896, context_id=None)


class StatefulGame(Game):
    """Games that have state."""

    __slots__ = ("_state",)

    def __init__(self, state: ConnectionState, **kwargs: Any):
        super().__init__(**kwargs)
        self._state = state

    async def clan(self) -> Clan:
        """Fetch this game's clan.

        This can be useful to get a Game's updates.

        .. code-block:: python3

            clan = await game.clan()
            async for update in clan.announcements().filter(
                lambda announcement: announcement.type in (steam.ClanEvent.MajorUpdate, steam.ClanEvent.SmallUpdate)
            ):
                ...  # do something with the update

        Raises
        ------
        :exc:`ValueError`
            This game has no associated clan.
        """

        id64 = await id64_from_url(self.url, self._state.http._session)
        if id64 is None:
            raise ValueError("Game has no associated clan")
        clan = await self._state.fetch_clan(id64)
        if clan is None:
            raise ValueError("Game has no associated clan")
        return clan

    async def player_count(self) -> int:
        """The games current player count."""
        return await self._state.fetch_game_player_count(self.id)

    async def friends_who_own(self) -> list[User]:
        """Fetch the users in your friend list who own this game."""
        id64s = await self._state.fetch_friends_who_own(self.id)
        return [self._state.get_user(id64) for id64 in id64s]  # type: ignore  # friends are always cached

    # async def fetch(self) -> Self & FetchedGame:  # TODO update signature to this when types.Intersection is done
    #     fetched = await self._state.client.fetch_game(self)
    #     return utils.update_class(fetched, copy.copy(self))

    async def fetch(self) -> FetchedGame:
        """Short hand for:

        .. code-block:: python3

            game = await client.fetch_game(game)
        """
        game = await self._state.client.fetch_game(self)
        if game is None:
            raise ValueError("Fetched game was not valid.")
        return game


class UserGame(StatefulGame):
    """Represents a Steam game fetched by :meth:`steam.User.games`

    Attributes
    ----------
    total_play_time
        The total time the game has been played for.
    icon_url
        The icon url of the game.
    logo_url
        The logo url of the game.
    """

    __slots__ = (
        "total_play_time",
        "icon_url",
        "logo_url",
        "_stats_visible",
    )

    name: str

    def __init__(self, state: ConnectionState, data: GameDict):
        super().__init__(state=state, id=data["appid"], name=data["name"])
        self.total_play_time: timedelta = timedelta(minutes=data["playtime_forever"])
        self.icon_url = (
            f"https://cdn.cloudflare.steamstatic.com/steamcommunity/public/images/apps/{self.id}/"
            f"{data['img_icon_url']}.jpg"
        )
        self.logo_url = (
            f"https://cdn.cloudflare.steamstatic.com/steamcommunity/public/images/apps/{self.id}/"
            f"{data['img_logo_url']}.jpg"
        )
        self._stats_visible = data.get("has_community_visible_stats", False)

    def has_visible_stats(self) -> bool:
        """Whether the game has publicly visible stats."""
        return self._stats_visible


class WishlistGame(StatefulGame):
    """Represents a Steam game fetched by :meth:`steam.User.wishlist`\\.

    Attributes
    ----------
    priority
        The priority of the game in the wishlist.
    added_at
        The time that the game was added to their wishlist.
    created_at
        The time the game was uploaded at.
    background_url
        The background URL of the game.
    rank
        The global rank of the game by popularity.
    review_status
        The review status of the game.
    score
        The score of the game out of ten.
    screenshots
        The screenshots of the game.
    tags
        The tags of the game.
    total_reviews
        The total number reviews for the game.
    type
        The type of the app.
    logo_url
        The logo url of the game.
    """

    __slots__ = (
        "priority",
        "added_at",
        "background_url",
        "created_at",
        "logo_url",
        "rank",
        "review_status",
        "score",
        "screenshots",
        "tags",
        "total_reviews",
        "type",
        "_free",
        "_on_linux",
        "_on_mac_os",
        "_on_windows",
    )

    name: str

    def __init__(self, state: ConnectionState, id: int | str, data: WishlistGameDict):
        super().__init__(state, id=id, name=data["name"])
        self.logo_url = data["capsule"]
        self.score = data["review_score"]
        self.total_reviews = int(data["reviews_total"].replace(",", ""))
        self.review_status = ReviewType[data["review_desc"].replace(" ", "")]
        self.created_at = datetime.utcfromtimestamp(int(data["release_date"]))
        self.type: str = data["type"]
        self.screenshots = [
            f"https://cdn.cloudflare.steamstatic.com/steam/apps/{self.id}/{screenshot_url}"
            for screenshot_url in data["screenshots"]
        ]
        self.added_at = datetime.utcfromtimestamp(data["added"])
        self.background_url = data["background"]
        self.tags = data["tags"]
        self.rank = data["rank"]
        self.priority = int(data["priority"])

        self._free = data["is_free_game"]
        self._on_windows = bool(data.get("win", False))
        self._on_mac_os = bool(data.get("mac", False))
        self._on_linux = bool(data.get("linux", False))

    def is_free(self) -> bool:
        """Whether the game is free to download."""
        return self._free

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux


class Movie:
    __slots__ = ("name", "id", "url", "created_at")

    def __init__(self, movie: dict[str, Any]):
        self.name: str = movie["name"]
        self.id: int = movie["id"]
        self.url: str = movie["mp4"]["max"]
        match = re.search(r"t=(\d+)", self.url)
        self.created_at = datetime.utcfromtimestamp(int(match.group(1))) if match else None

    def __repr__(self) -> str:
        attrs = ("name", "id", "url", "created_at")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Movie {' '.join(resolved)}>"


class FetchedGame(StatefulGame):
    """Represents a Steam game fetched by :meth:`steam.Client.fetch_game`\\.

    Attributes
    ----------
    created_at
        The time the game was uploaded at.
    background_url
        The background URL of the game.
    type
        The type of the app.
    logo_url
        The logo URL of the game.
    dlc
        The game's downloadable content.
    website_url
        The website URL of the game.
    developers
        The developers of the game.
    publishers
        The publishers of the game.
    description
        The short description of the game.
    full_description
        The full description of the game.
    movies
        A list of the game's movies, each of which has ``name``\\, ``id``\\, ``url`` and optional
        ``created_at`` attributes.
    """

    __slots__ = (
        "logo_url",
        "background_url",
        "created_at",
        "type",
        "dlc",
        "website_url",
        "developers",
        "publishers",
        "description",
        "full_description",
        "movies",
        "_free",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )

    title: str

    def __init__(self, state: ConnectionState, data: FetchedGameDict):
        super().__init__(state, id=data["steam_appid"], name=data["name"])
        self.logo_url = data["header_image"]
        self.background_url = data["background"]
        self.created_at = (
            datetime.strptime(data["release_date"]["date"], "%d %b, %Y") if data["release_date"]["date"] else None
        )
        self.type = data["type"]

        self.dlc = [Game(id=dlc_id) for dlc_id in data["dlc"]] if "dlc" in data else None
        self.website_url = data.get("website")
        self.developers = data["developers"]
        self.publishers = data["publishers"]
        self.description = data["short_description"]
        self.full_description = data["detailed_description"]

        self.movies = [Movie(movie) for movie in data["movies"]] if "movies" in data else None

        self._free = data["is_free"]
        self._on_windows = bool(data["platforms"].get("windows", False))
        self._on_mac_os = bool(data["platforms"].get("mac", False))
        self._on_linux = bool(data["platforms"].get("linux", False))

    def is_free(self) -> bool:
        """Whether the game is free to download."""
        return self._free

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux



from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class UpdateRequest(betterproto.Message):
    recommendationid: int = betterproto.uint64_field(1)
    review_text: str = betterproto.string_field(2)
    voted_up: bool = betterproto.bool_field(3)
    is_public: bool = betterproto.bool_field(4)
    language: str = betterproto.string_field(5)
    is_in_early_access: bool = betterproto.bool_field(6)
    received_compensation: bool = betterproto.bool_field(7)
    comments_disabled: bool = betterproto.bool_field(8)


@dataclass(eq=False, repr=False)
class UpdateResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GetIndividualRecommendationsRequest(betterproto.Message):
    requests: List["GetIndividualRecommendationsRequestRecommendationRequest"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetIndividualRecommendationsRequestRecommendationRequest(betterproto.Message):
    steamid: int = betterproto.uint64_field(1)
    appid: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class GetIndividualRecommendationsResponse(betterproto.Message):
    recommendations: List["RecommendationDetails"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class RecommendationDetails(betterproto.Message):
    recommendationid: int = betterproto.uint64_field(1)
    steamid: int = betterproto.uint64_field(2)
    appid: int = betterproto.uint32_field(3)
    review: str = betterproto.string_field(4)
    time_created: int = betterproto.uint32_field(5)
    time_updated: int = betterproto.uint32_field(6)
    votes_up: int = betterproto.uint32_field(7)
    votes_down: int = betterproto.uint32_field(8)
    vote_score: float = betterproto.float_field(9)
    language: str = betterproto.string_field(10)
    comment_count: int = betterproto.uint32_field(11)
    voted_up: bool = betterproto.bool_field(12)
    is_public: bool = betterproto.bool_field(13)
    moderator_hidden: bool = betterproto.bool_field(14)
    flagged_by_developer: int = betterproto.int32_field(15)
    report_score: int = betterproto.uint32_field(16)
    steamid_moderator: int = betterproto.uint64_field(17)
    steamid_developer: int = betterproto.uint64_field(18)
    steamid_dev_responder: int = betterproto.uint64_field(19)
    developer_response: str = betterproto.string_field(20)
    time_developer_responded: int = betterproto.uint32_field(21)
    developer_flag_cleared: bool = betterproto.bool_field(22)
    written_during_early_access: bool = betterproto.bool_field(23)
    votes_funny: int = betterproto.uint32_field(24)
    received_compensation: bool = betterproto.bool_field(25)
    unverified_purchase: bool = betterproto.bool_field(26)
    review_quality: int = betterproto.int32_field(27)
    weighted_vote_score: float = betterproto.float_field(28)
    moderation_note: str = betterproto.string_field(29)
    payment_method: int = betterproto.int32_field(30)
    playtime_2weeks: int = betterproto.int32_field(31)
    playtime_forever: int = betterproto.int32_field(32)
    last_playtime: int = betterproto.int32_field(33)
    comments_disabled: bool = betterproto.bool_field(34)
    playtime_at_review: int = betterproto.int32_field(35)
    approved_for_china: bool = betterproto.bool_field(36)
    ban_check_result: int = betterproto.int32_field(37)
    refunded: bool = betterproto.bool_field(38)
    account_score_spend: int = betterproto.int32_field(39)
    reactions: List["RecommendationLoyaltyReaction"] = betterproto.message_field(40)
    ipaddress: str = betterproto.string_field(41)


@dataclass(eq=False, repr=False)
class RecommendationLoyaltyReaction(betterproto.Message):
    reaction_type: int = betterproto.uint32_field(1)
    count: int = betterproto.uint32_field(2)
